from libqtile.utils import acall_process
from libqtile.widget import base


class GenPollText(base.BackgroundPoll):
    """A generic text widget that polls using poll function to get the text

    Widget requirements: aiohttp_.

    .. _aiohttp: https://pypi.org/project/aiohttp/
    """

    defaults = [
        ("func", None, "Poll Function"),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(GenPollText.defaults)

    def poll(self):
        if not self.func:
            return "You need a poll function"
        return self.func()


class GenPollCommand(base.BackgroundPoll):
    """A generic text widget to display output from scripts or shell commands"""

    defaults = [
        ("update_interval", 60, "update time in seconds"),
        ("cmd", None, "command line as a string or list of arguments to execute"),
        ("shell", False, "run command through shell to enable piping and shell expansion"),
        ("parse", None, "Function to parse output of command"),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(GenPollCommand.defaults)

    def _configure(self, qtile, bar):
        base.BackgroundPoll._configure(self, qtile, bar)
        self.add_callbacks({"Button1": self.force_update})

    async def apoll(self):
        out = await acall_process(self.cmd, self.shell)
        if self.parse:
            return self.parse(out)

        return out.strip()



from libqtile import bar
from libqtile.widget import base


class Spacer(base._Widget):
    """Just an empty space on the bar

    Often used with length equal to bar.STRETCH to push bar widgets to the
    right or bottom edge of the screen.

    Parameters
    ==========
    length :
        Length of the widget.  Can be either ``bar.STRETCH`` or a length in
        pixels.
    width :
        DEPRECATED, same as ``length``.
    """

    orientations = base.ORIENTATION_BOTH
    defaults = [("background", None, "Widget background color")]

    def __init__(self, length=bar.STRETCH, **config):
        """ """
        base._Widget.__init__(self, length, **config)
        self.add_defaults(Spacer.defaults)

    def draw(self):
        if self.length > 0:
            self.drawer.clear(self.background or self.bar.background)
            self.draw_at_default_position()



from __future__ import annotations

import math
from collections import namedtuple
from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import _SimpleLayoutBase

if TYPE_CHECKING:
    from typing import Any, Self

    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect
    from libqtile.group import _Group


class MonadTall(_SimpleLayoutBase):
    """Emulate the behavior of XMonad's default tiling scheme.

    Main-Pane:

    A main pane that contains a single window takes up a vertical portion of
    the screen_rect based on the ratio setting. This ratio can be adjusted with
    the ``grow_main`` and ``shrink_main`` or, while the main pane is in
    focus, ``grow`` and ``shrink``. You may also set the ratio directly
    with ``set_ratio``.

    ::

        ---------------------
        |            |      |
        |            |      |
        |            |      |
        |            |      |
        |            |      |
        |            |      |
        ---------------------

    Using the ``flip`` method will switch which horizontal side the main
    pane will occupy. The main pane is considered the "top" of the stack.

    ::

        ---------------------
        |      |            |
        |      |            |
        |      |            |
        |      |            |
        |      |            |
        |      |            |
        ---------------------

    Secondary-panes:

    Occupying the rest of the screen_rect are one or more secondary panes.  The
    secondary panes will share the vertical space of the screen_rect however
    they can be resized at will with the ``grow`` and ``shrink``
    methods.  The other secondary panes will adjust their sizes to smoothly fill
    all of the space.

    ::

        ---------------------          ---------------------
        |            |      |          |            |______|
        |            |______|          |            |      |
        |            |      |          |            |      |
        |            |______|          |            |      |
        |            |      |          |            |______|
        |            |      |          |            |      |
        ---------------------          ---------------------

    Panes can be moved with the ``shuffle_up`` and ``shuffle_down``
    methods. As mentioned the main pane is considered the top of the stack;
    moving up is counter-clockwise and moving down is clockwise.

    ::

        ---------------------          ---------------------
        |            |      |          |            |      |
        |            |______|          |            |Focus |
        |            |      |          |            |      |
        |            |______|          |            |______|
        |            |      |          |            |______|
        |            |      |          |            |      |
        ---------------------          ---------------------

    Setting ``auto_maximize`` will cause the focused secondary pane to be
    automatically maximized on focus. The non-maximized panes will shrink to
    the height specified by ``min_secondary_size``.

    The opposite is true if the layout is "flipped".

    ::

        ---------------------          ---------------------
        |            |  2   |          |   2   |           |
        |            |______|          |_______|           |
        |            |  3   |          |   3   |           |
        |     1      |______|          |_______|     1     |
        |            |  4   |          |   4   |           |
        |            |      |          |       |           |
        ---------------------          ---------------------


    Normalizing/Resetting:

    To restore all secondary client windows to their default size ratios
    use the ``normalize`` method.

    To reset all client windows to their default sizes, including the primary
    window, use the ``reset`` method.

    Maximizing:

    To toggle a client window between its minimum and maximum sizes
    simply use the ``maximize`` on a focused client.

    Suggested Bindings::

        Key([modkey], "h", lazy.layout.left()),
        Key([modkey], "l", lazy.layout.right()),
        Key([modkey], "j", lazy.layout.down()),
        Key([modkey], "k", lazy.layout.up()),
        Key([modkey, "shift"], "h", lazy.layout.swap_left()),
        Key([modkey, "shift"], "l", lazy.layout.swap_right()),
        Key([modkey, "shift"], "j", lazy.layout.shuffle_down()),
        Key([modkey, "shift"], "k", lazy.layout.shuffle_up()),
        Key([modkey], "i", lazy.layout.grow()),
        Key([modkey], "m", lazy.layout.shrink()),
        Key([modkey], "n", lazy.layout.reset()),
        Key([modkey, "shift"], "n", lazy.layout.normalize()),
        Key([modkey], "o", lazy.layout.maximize()),
        Key([modkey, "shift"], "s", lazy.layout.toggle_auto_maximize()),
        Key([modkey, "shift"], "space", lazy.layout.flip()),
    """

    _left = 0
    _right = 1

    defaults = [
        ("auto_maximize", False, "Maximize secondary windows on focus."),
        ("border_focus", "#ff0000", "Border colour(s) for the focused window."),
        ("border_normal", "#000000", "Border colour(s) for un-focused windows."),
        ("border_width", 2, "Border width."),
        ("single_border_width", None, "Border width for single window"),
        ("single_margin", None, "Margin size for single window"),
        ("margin", 0, "Margin of the layout"),
        (
            "ratio",
            0.5,
            "The percent of the screen-space the master pane should occupy by default.",
        ),
        (
            "min_ratio",
            0.25,
            "The percent of the screen-space the master pane should occupy at minimum.",
        ),
        (
            "max_ratio",
            0.75,
            "The percent of the screen-space the master pane should occupy at maximum.",
        ),
        ("min_secondary_size", 85, "minimum size in pixel for a secondary pane window "),
        (
            "align",
            _left,
            "Which side master plane will be placed "
            "(one of ``MonadTall._left`` or ``MonadTall._right``)",
        ),
        ("change_ratio", 0.05, "Resize ratio"),
        ("change_size", 20, "Resize change in pixels"),
        (
            "new_client_position",
            "after_current",
            "Place new windows: "
            " after_current - after the active window."
            " before_current - before the active window,"
            " top - at the top of the stack,"
            " bottom - at the bottom of the stack,",
        ),
    ]

    def __init__(self, **config):
        _SimpleLayoutBase.__init__(self, **config)
        self.add_defaults(MonadTall.defaults)
        if self.single_border_width is None:
            self.single_border_width = self.border_width
        if self.single_margin is None:
            self.single_margin = self.margin
        self.relative_sizes = []
        self._screen_rect = None
        self.default_ratio = self.ratio

    # screen_rect is a property as the MonadThreeCol layout needs to perform
    # additional actions when the attribute is modified
    @property
    def screen_rect(self):
        return self._screen_rect

    @screen_rect.setter
    def screen_rect(self, value):
        self._screen_rect = value

    @property
    def focused(self):
        return self.clients.current_index

    def _get_relative_size_from_absolute(self, absolute_size):
        return absolute_size / self.screen_rect.height

    def _get_absolute_size_from_relative(self, relative_size):
        return int(relative_size * self.screen_rect.height)

    def clone(self, group: _Group) -> Self:
        "Clone layout for other groups"
        c = _SimpleLayoutBase.clone(self, group)
        c.relative_sizes = []
        c.screen_rect = group.screen.get_rect() if group.screen else None
        c.ratio = self.ratio
        c.align = self.align
        return c

    def add_client(self, client: Window) -> None:  # type: ignore[override]
        "Add client to layout"
        self.clients.add_client(client, client_position=self.new_client_position)
        self.do_normalize = True

    def focus(self, client):
        super().focus(client)
        # Only maximize the window in the secondary pane when focus is *not* in
        # the main pane. Doing so in the main pane causes the last secondary
        # window to always be in focus when switching from secondary -> main.
        if self.focused != 0:
            self._maximize_focused_secondary()

    def remove(self, client: Window) -> Window | None:
        "Remove client from layout"
        p = super().remove(client)
        self.do_normalize = True
        # When auto_maximize is set and the user closes the first (topmost)
        # secondary window, focus goes back to the main window. In this case,
        # we WANT to force redraw of the windows in the secondary pane so we
        # get a maximized topmost window again.
        if self.auto_maximize and self.focused == 0 and len(self.clients) > 2:
            # This will also trigger secondary maximization, if needed.
            self.focus(self.clients[1])
        return p

    @expose_command()
    def set_ratio(self, ratio):
        "Directly set the main pane ratio"
        ratio = min(self.max_ratio, ratio)
        self.ratio = max(self.min_ratio, ratio)
        self.group.layout_all()

    @expose_command()
    def normalize(self, redraw=True):
        "Evenly distribute screen-space among secondary clients"
        n = len(self.clients) - 1  # exclude main client, 0
        # if secondary clients exist
        if n > 0 and self.screen_rect is not None:
            self.relative_sizes = [1.0 / n] * n
        # reset main pane ratio
        if redraw:
            self.group.layout_all()
        self.do_normalize = False

    @expose_command()
    def reset(self, ratio=None, redraw=True):
        "Reset Layout."
        self.ratio = ratio or self.default_ratio
        if self.align == self._right:
            self.align = self._left
        self.normalize(redraw)

    @expose_command()
    def toggle_auto_maximize(self):
        "Toggle auto maximize secondary window on focus."
        self.auto_maximize = not self.auto_maximize
        self.normalize(True)
        if self.focused != 0:
            self._maximize_focused_secondary()

    def _maximize_main(self):
        "Toggle the main pane between min and max size"
        if self.ratio <= 0.5 * (self.max_ratio + self.min_ratio):
            self.ratio = self.max_ratio
        else:
            self.ratio = self.min_ratio
        self.group.layout_all()

    def _maximize_secondary(self):
        "Toggle the focused secondary pane between min and max size"
        n = len(self.clients) - 2  # total shrinking clients
        # total size of collapsed secondaries
        collapsed_size = self.min_secondary_size * n
        nidx = self.focused - 1  # focused size index
        # total height of maximized secondary
        maxed_size = self.group.screen.dheight - collapsed_size
        # if maximized or nearly maximized
        if (
            abs(self._get_absolute_size_from_relative(self.relative_sizes[nidx]) - maxed_size)
            < self.change_size
        ):
            # minimize
            self._shrink_secondary(
                self._get_absolute_size_from_relative(self.relative_sizes[nidx])
                - self.min_secondary_size
            )
        # otherwise maximize
        else:
            self._grow_secondary(maxed_size)

    def _maximize_focused_secondary(self):
        "Maximize the 'non-maximized' focused secondary pane"

        # Return immediately if no self.group.screen
        # (this may happen when moving windows across screens.)
        if self.group.screen is None:
            return

        # If auto_maximize is off, return immediately.
        if not self.auto_maximize:
            return

        # if we have 1 or 2 panes, do nothing.
        if len(self.clients) < 3:
            return

        # Recalculate relative_sizes
        self.normalize(redraw=False)
        if len(self.relative_sizes) == 0:
            return

        # If the focused window (self.focused) is 0 (main pane), adjust
        # focused to work directly on the secondary pane windows.
        focused = max(1, self.focused)

        n = len(self.clients) - 2  # total shrinking clients
        # total size of collapsed secondaries
        collapsed_size = self.min_secondary_size * n
        nidx = max(0, focused - 1)  # focused size index
        # total height of maximized secondary
        maxed_size = self.group.screen.dheight - collapsed_size

        # Maximize if window is not already maximized.
        if (
            abs(self._get_absolute_size_from_relative(self.relative_sizes[nidx]) - maxed_size)
            >= self.change_size
        ):
            self._grow_secondary(maxed_size)
            self.group.layout_all()

    @expose_command()
    def maximize(self):
        "Grow the currently focused client to the max size"
        # if we have 1 or 2 panes or main pane is focused
        if len(self.clients) < 3 or self.focused == 0:
            self._maximize_main()
        # secondary is focused
        else:
            self._maximize_secondary()
        self.group.layout_all()

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        "Position client based on order and sizes"
        self.screen_rect = screen_rect

        # if no sizes or normalize flag is set, normalize
        if not self.relative_sizes or self.do_normalize:
            self.normalize(False)

        # if client not in this layout
        if not self.clients or client not in self.clients:
            client.hide()
            return

        # determine focus border-color
        if client.has_focus:
            px = self.border_focus
        else:
            px = self.border_normal

        # single client - fullscreen
        if len(self.clients) == 1:
            client.place(
                self.screen_rect.x,
                self.screen_rect.y,
                self.screen_rect.width - 2 * self.single_border_width,
                self.screen_rect.height - 2 * self.single_border_width,
                self.single_border_width,
                px,
                margin=self.single_margin,
            )
            client.unhide()
            return
        cidx = self.clients.index(client)
        self._configure_specific(client, screen_rect, px, cidx)
        client.unhide()

    def _configure_specific(self, client, screen_rect, px, cidx):
        """Specific configuration for xmonad tall."""
        self.screen_rect = screen_rect

        # calculate main/secondary pane size
        width_main = int(self.screen_rect.width * self.ratio)
        width_shared = self.screen_rect.width - width_main

        # calculate client's x offset
        if self.align == self._left:  # left or up orientation
            if cidx == 0:
                # main client
                xpos = self.screen_rect.x
            else:
                # secondary client
                xpos = self.screen_rect.x + width_main
        else:  # right or down orientation
            if cidx == 0:
                # main client
                xpos = self.screen_rect.x + width_shared - self.margin
            else:
                # secondary client
                xpos = self.screen_rect.x

        # calculate client height and place
        if cidx > 0:
            # secondary client
            width = width_shared - 2 * self.border_width
            # ypos is the sum of all clients above it
            ypos = self.screen_rect.y + self._get_absolute_size_from_relative(
                sum(self.relative_sizes[: cidx - 1])
            )
            # get height from precalculated height list
            height = self._get_absolute_size_from_relative(self.relative_sizes[cidx - 1])
            # fix double margin
            if cidx > 1:
                ypos -= self.margin
                height += self.margin
            # place client based on calculated dimensions
            client.place(
                xpos,
                ypos,
                width,
                height - 2 * self.border_width,
                self.border_width,
                px,
                margin=self.margin,
            )
        else:
            # main client
            client.place(
                xpos,
                self.screen_rect.y,
                width_main,
                self.screen_rect.height,
                self.border_width,
                px,
                margin=[
                    self.margin,
                    2 * self.border_width,
                    self.margin + 2 * self.border_width,
                    self.margin,
                ],
            )

    @expose_command()
    def info(self) -> dict[str, Any]:
        d = _SimpleLayoutBase.info(self)
        d.update(
            dict(
                main=d["clients"][0] if self.clients else None,
                secondary=d["clients"][1::] if self.clients else [],
            )
        )
        return d

    def get_shrink_margin(self, cidx):
        "Return how many remaining pixels a client can shrink"
        return max(
            0,
            self._get_absolute_size_from_relative(self.relative_sizes[cidx])
            - self.min_secondary_size,
        )

    def _shrink(self, cidx, amt):
        """Reduce the size of a client

        Will only shrink the client until it reaches the configured minimum
        size. Any amount that was prevented in the resize is returned.
        """
        # get max resizable amount
        margin = self.get_shrink_margin(cidx)
        if amt > margin:  # too much
            self.relative_sizes[cidx] -= self._get_relative_size_from_absolute(margin)
            return amt - margin
        else:
            self.relative_sizes[cidx] -= self._get_relative_size_from_absolute(amt)
            return 0

    def shrink_up(self, cidx, amt):
        """Shrink the window up

        Will shrink all secondary clients above the specified index in order.
        Each client will attempt to shrink as much as it is able before the
        next client is resized.

        Any amount that was unable to be applied to the clients is returned.
        """
        left = amt  # track unused shrink amount
        # for each client before specified index
        for idx in range(cidx):
            # shrink by whatever is left-over of original amount
            left -= left - self._shrink(idx, left)
        # return unused shrink amount
        return left

    def shrink_up_shared(self, cidx, amt):
        """Shrink the shared space

        Will shrink all secondary clients above the specified index by an equal
        share of the provided amount. After applying the shared amount to all
        affected clients, any amount left over will be applied in a non-equal
        manner with ``shrink_up``.

        Any amount that was unable to be applied to the clients is returned.
        """
        # split shrink amount among number of clients
        per_amt = amt / cidx
        left = amt  # track unused shrink amount
        # for each client before specified index
        for idx in range(cidx):
            # shrink by equal amount and track left-over
            left -= per_amt - self._shrink(idx, per_amt)
        # apply non-equal shrinkage secondary pass
        # in order to use up any left over shrink amounts
        left = self.shrink_up(cidx, left)
        # return whatever could not be applied
        return left

    def shrink_down(self, cidx, amt):
        """Shrink current window down

        Will shrink all secondary clients below the specified index in order.
        Each client will attempt to shrink as much as it is able before the
        next client is resized.

        Any amount that was unable to be applied to the clients is returned.
        """
        left = amt  # track unused shrink amount
        # for each client after specified index
        for idx in range(cidx + 1, len(self.relative_sizes)):
            # shrink by current total left-over amount
            left -= left - self._shrink(idx, left)
        # return unused shrink amount
        return left

    def shrink_down_shared(self, cidx, amt):
        """Shrink secondary clients

        Will shrink all secondary clients below the specified index by an equal
        share of the provided amount. After applying the shared amount to all
        affected clients, any amount left over will be applied in a non-equal
        manner with ``shrink_down``.

        Any amount that was unable to be applied to the clients is returned.
        """
        # split shrink amount among number of clients
        per_amt = amt / (len(self.relative_sizes) - 1 - cidx)
        left = amt  # track unused shrink amount
        # for each client after specified index
        for idx in range(cidx + 1, len(self.relative_sizes)):
            # shrink by equal amount and track left-over
            left -= per_amt - self._shrink(idx, per_amt)
        # apply non-equal shrinkage secondary pass
        # in order to use up any left over shrink amounts
        left = self.shrink_down(cidx, left)
        # return whatever could not be applied
        return left

    def _grow_main(self, amt):
        """Will grow the client that is currently in the main pane"""
        self.ratio += amt
        self.ratio = min(self.max_ratio, self.ratio)

    def _grow_solo_secondary(self, amt):
        """Will grow the solitary client in the secondary pane"""
        self.ratio -= amt
        self.ratio = max(self.min_ratio, self.ratio)

    def _grow_secondary(self, amt):
        """Will grow the focused client in the secondary pane"""
        half_change_size = amt / 2
        # track unshrinkable amounts
        left = amt
        # first secondary (top)
        if self.focused == 1:
            # only shrink downwards
            left -= amt - self.shrink_down_shared(0, amt)
        # last secondary (bottom)
        elif self.focused == len(self.clients) - 1:
            # only shrink upwards
            left -= amt - self.shrink_up(len(self.relative_sizes) - 1, amt)
        # middle secondary
        else:
            # get size index
            idx = self.focused - 1
            # shrink up and down
            left -= half_change_size - self.shrink_up_shared(idx, half_change_size)
            left -= half_change_size - self.shrink_down_shared(idx, half_change_size)
            left -= half_change_size - self.shrink_up_shared(idx, half_change_size)
            left -= half_change_size - self.shrink_down_shared(idx, half_change_size)
        # calculate how much shrinkage took place
        diff = amt - left
        # grow client by diff amount
        self.relative_sizes[self.focused - 1] += self._get_relative_size_from_absolute(diff)

    @expose_command()
    def grow(self):
        """Grow current window

        Will grow the currently focused client reducing the size of those
        around it. Growing will stop when no other secondary clients can reduce
        their size any further.
        """
        if self.focused == 0:
            self._grow_main(self.change_ratio)
        elif len(self.clients) == 2:
            self._grow_solo_secondary(self.change_ratio)
        else:
            self._grow_secondary(self.change_size)
        self.group.layout_all()

    @expose_command()
    def grow_main(self):
        """Grow main pane

        Will grow the main pane, reducing the size of clients in the secondary
        pane.
        """
        self._grow_main(self.change_ratio)
        self.group.layout_all()

    @expose_command()
    def shrink_main(self):
        """Shrink main pane

        Will shrink the main pane, increasing the size of clients in the
        secondary pane.
        """
        self._shrink_main(self.change_ratio)
        self.group.layout_all()

    def _grow(self, cidx, amt):
        "Grow secondary client by specified amount"
        self.relative_sizes[cidx] += self._get_relative_size_from_absolute(amt)

    def grow_up_shared(self, cidx, amt):
        """Grow higher secondary clients

        Will grow all secondary clients above the specified index by an equal
        share of the provided amount.
        """
        # split grow amount among number of clients
        per_amt = amt / cidx
        for idx in range(cidx):
            self._grow(idx, per_amt)

    def grow_down_shared(self, cidx, amt):
        """Grow lower secondary clients

        Will grow all secondary clients below the specified index by an equal
        share of the provided amount.
        """
        # split grow amount among number of clients
        per_amt = amt / (len(self.relative_sizes) - 1 - cidx)
        for idx in range(cidx + 1, len(self.relative_sizes)):
            self._grow(idx, per_amt)

    def _shrink_main(self, amt):
        """Will shrink the client that currently in the main pane"""
        self.ratio -= amt
        self.ratio = max(self.min_ratio, self.ratio)

    def _shrink_solo_secondary(self, amt):
        """Will shrink the solitary client in the secondary pane"""
        self.ratio += amt
        self.ratio = min(self.max_ratio, self.ratio)

    def _shrink_secondary(self, amt):
        """Will shrink the focused client in the secondary pane"""
        # get focused client
        client = self.clients[self.focused]

        # get default change size
        change = amt

        # get left-over height after change
        left = client.height - amt
        # if change would violate min_secondary_size
        if left < self.min_secondary_size:
            # just reduce to min_secondary_size
            change = client.height - self.min_secondary_size

        # calculate half of that change
        half_change = change / 2

        # first secondary (top)
        if self.focused == 1:
            # only grow downwards
            self.grow_down_shared(0, change)
        # last secondary (bottom)
        elif self.focused == len(self.clients) - 1:
            # only grow upwards
            self.grow_up_shared(len(self.relative_sizes) - 1, change)
        # middle secondary
        else:
            idx = self.focused - 1
            # grow up and down
            self.grow_up_shared(idx, half_change)
            self.grow_down_shared(idx, half_change)
        # shrink client by total change
        self.relative_sizes[self.focused - 1] -= self._get_relative_size_from_absolute(change)

    @expose_command()
    def next(self) -> None:
        _SimpleLayoutBase.next(self)

    @expose_command()
    def previous(self) -> None:
        _SimpleLayoutBase.previous(self)

    @expose_command()
    def shrink(self):
        """Shrink current window

        Will shrink the currently focused client reducing the size of those
        around it. Shrinking will stop when the client has reached the minimum
        size.
        """
        if self.focused == 0:
            self._shrink_main(self.change_ratio)
        elif len(self.clients) == 2:
            self._shrink_solo_secondary(self.change_ratio)
        else:
            self._shrink_secondary(self.change_size)
        self.group.layout_all()

    @expose_command()
    def shuffle_up(self):
        """Shuffle the client up the stack"""
        self.clients.shuffle_up()
        self.group.layout_all()
        self.group.focus(self.clients.current_client)

    @expose_command()
    def shuffle_down(self):
        """Shuffle the client down the stack"""
        self.clients.shuffle_down()
        self.group.layout_all()
        self.group.focus(self.clients[self.focused])

    @expose_command()
    def flip(self):
        """Flip the layout horizontally"""
        self.align = self._left if self.align == self._right else self._right
        self.group.layout_all()

    def _get_closest(self, x, y, clients):
        """Get closest window to a point x,y"""
        target = min(
            clients,
            key=lambda c: math.hypot(c.x - x, c.y - y),
            default=self.clients.current_client,
        )
        return target

    @expose_command()
    def swap(self, window1: Window, window2: Window) -> None:
        """Swap two windows"""
        _SimpleLayoutBase.swap(self, window1, window2)

    @expose_command("shuffle_left")
    def swap_left(self):
        """Swap current window with closest window to the left"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["x"] < x]
        target = self._get_closest(x, y, candidates)
        self.swap(win, target)

    @expose_command("shuffle_right")
    def swap_right(self):
        """Swap current window with closest window to the right"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["x"] > x]
        target = self._get_closest(x, y, candidates)
        self.swap(win, target)

    @expose_command()
    def swap_main(self):
        """Swap current window to main pane"""
        if self.align == self._left:
            self.swap_left()
        elif self.align == self._right:
            self.swap_right()

    @expose_command()
    def left(self):
        """Focus on the closest window to the left of the current window"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["x"] < x]
        self.clients.current_client = self._get_closest(x, y, candidates)
        self.group.focus(self.clients.current_client)

    @expose_command()
    def right(self):
        """Focus on the closest window to the right of the current window"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["x"] > x]
        self.clients.current_client = self._get_closest(x, y, candidates)
        self.group.focus(self.clients.current_client)

    @expose_command()
    def up(self):
        """Focus on the closest window above the current window"""
        self.previous()

    @expose_command()
    def down(self):
        """Focus on the closest window below the current window"""
        self.next()


class MonadWide(MonadTall):
    """Emulate the behavior of XMonad's horizontal tiling scheme.

    This layout attempts to emulate the behavior of XMonad wide
    tiling scheme.

    Main-Pane:

    A main pane that contains a single window takes up a horizontal
    portion of the screen_rect based on the ratio setting. This ratio can be
    adjusted with the ``grow_main`` and ``shrink_main`` or,
    while the main pane is in focus, ``grow`` and ``shrink``.

    ::

        ---------------------
        |                   |
        |                   |
        |                   |
        |___________________|
        |                   |
        |                   |
        ---------------------

    Using the ``flip`` method will switch which vertical side the
    main pane will occupy. The main pane is considered the "top" of
    the stack.

    ::

        ---------------------
        |                   |
        |___________________|
        |                   |
        |                   |
        |                   |
        |                   |
        ---------------------

    Secondary-panes:

    Occupying the rest of the screen_rect are one or more secondary panes.
    The secondary panes will share the horizontal space of the screen_rect
    however they can be resized at will with the ``grow`` and
    ``shrink`` methods. The other secondary panes will adjust their
    sizes to smoothly fill all of the space.

    ::

        ---------------------          ---------------------
        |                   |          |                   |
        |                   |          |                   |
        |                   |          |                   |
        |___________________|          |___________________|
        |     |       |     |          |   |           |   |
        |     |       |     |          |   |           |   |
        ---------------------          ---------------------

    Panes can be moved with the ``shuffle_up`` and ``shuffle_down``
    methods. As mentioned the main pane is considered the top of the
    stack; moving up is counter-clockwise and moving down is clockwise.

    The opposite is true if the layout is "flipped".

    ::

        ---------------------          ---------------------
        |                   |          |  2  |   3   |  4  |
        |         1         |          |_____|_______|_____|
        |                   |          |                   |
        |___________________|          |                   |
        |     |       |     |          |        1          |
        |  2  |   3   |  4  |          |                   |
        ---------------------          ---------------------

    Normalizing/Resetting:

    To restore all secondary client windows to their default size ratios
    use the ``normalize`` method.

    To reset all client windows to their default sizes, including the primary
    window, use the ``reset`` method.


    Maximizing:

    To toggle a client window between its minimum and maximum sizes
    simply use the ``maximize`` on a focused client.

    Suggested Bindings::

        Key([modkey], "h", lazy.layout.left()),
        Key([modkey], "l", lazy.layout.right()),
        Key([modkey], "j", lazy.layout.down()),
        Key([modkey], "k", lazy.layout.up()),
        Key([modkey, "shift"], "h", lazy.layout.swap_left()),
        Key([modkey, "shift"], "l", lazy.layout.swap_right()),
        Key([modkey, "shift"], "j", lazy.layout.shuffle_down()),
        Key([modkey, "shift"], "k", lazy.layout.shuffle_up()),
        Key([modkey], "i", lazy.layout.grow()),
        Key([modkey], "m", lazy.layout.shrink()),
        Key([modkey], "n", lazy.layout.reset()),
        Key([modkey, "shift"], "n", lazy.layout.normalize()),
        Key([modkey], "o", lazy.layout.maximize()),
        Key([modkey, "shift"], "space", lazy.layout.flip()),
    """

    _up = 0
    _down = 1

    def _get_relative_size_from_absolute(self, absolute_size):
        return absolute_size / self.screen_rect.width

    def _get_absolute_size_from_relative(self, relative_size):
        return int(relative_size * self.screen_rect.width)

    def _maximize_secondary(self):
        """Toggle the focused secondary pane between min and max size."""
        n = len(self.clients) - 2  # total shrinking clients
        # total size of collapsed secondaries
        collapsed_size = self.min_secondary_size * n
        nidx = self.focused - 1  # focused size index
        # total width of maximized secondary
        maxed_size = self.screen_rect.width - collapsed_size
        # if maximized or nearly maximized
        if (
            abs(self._get_absolute_size_from_relative(self.relative_sizes[nidx]) - maxed_size)
            < self.change_size
        ):
            # minimize
            self._shrink_secondary(
                self._get_absolute_size_from_relative(self.relative_sizes[nidx])
                - self.min_secondary_size
            )
        # otherwise maximize
        else:
            self._grow_secondary(maxed_size)

    def _configure_specific(self, client, screen_rect, px, cidx):
        """Specific configuration for xmonad wide."""
        self.screen_rect = screen_rect

        # calculate main/secondary column widths
        height_main = int(self.screen_rect.height * self.ratio)
        height_shared = self.screen_rect.height - height_main

        # calculate client's x offset
        if self.align == self._up:  # up orientation
            if cidx == 0:
                # main client
                ypos = self.screen_rect.y
            else:
                # secondary client
                ypos = self.screen_rect.y + height_main
        else:  # right or down orientation
            if cidx == 0:
                # main client
                ypos = self.screen_rect.y + height_shared - self.margin
            else:
                # secondary client
                ypos = self.screen_rect.y

        # calculate client height and place
        if cidx > 0:
            # secondary client
            height = height_shared - 2 * self.border_width
            # xpos is the sum of all clients left of it
            xpos = self.screen_rect.x + self._get_absolute_size_from_relative(
                sum(self.relative_sizes[: cidx - 1])
            )
            # get width from precalculated width list
            width = self._get_absolute_size_from_relative(self.relative_sizes[cidx - 1])
            # fix double margin
            if cidx > 1:
                xpos -= self.margin
                width += self.margin
            # place client based on calculated dimensions
            client.place(
                xpos,
                ypos,
                width - 2 * self.border_width,
                height,
                self.border_width,
                px,
                margin=self.margin,
            )
        else:
            # main client
            client.place(
                self.screen_rect.x,
                ypos,
                self.screen_rect.width,
                height_main,
                self.border_width,
                px,
                margin=[
                    self.margin,
                    self.margin + 2 * self.border_width,
                    2 * self.border_width,
                    self.margin,
                ],
            )

    def _shrink_secondary(self, amt):
        """Will shrink the focused client in the secondary pane"""
        # get focused client
        client = self.clients[self.focused]

        # get default change size
        change = amt

        # get left-over height after change
        left = client.width - amt
        # if change would violate min_secondary_size
        if left < self.min_secondary_size:
            # just reduce to min_secondary_size
            change = client.width - self.min_secondary_size

        # calculate half of that change
        half_change = change / 2

        # first secondary (top)
        if self.focused == 1:
            # only grow downwards
            self.grow_down_shared(0, change)
        # last secondary (bottom)
        elif self.focused == len(self.clients) - 1:
            # only grow upwards
            self.grow_up_shared(len(self.relative_sizes) - 1, change)
        # middle secondary
        else:
            idx = self.focused - 1
            # grow up and down
            self.grow_up_shared(idx, half_change)
            self.grow_down_shared(idx, half_change)
        # shrink client by total change
        self.relative_sizes[self.focused - 1] -= self._get_relative_size_from_absolute(change)

    @expose_command()
    def swap_left(self):
        """Swap current window with closest window to the down"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients.clients if c.info()["y"] > y]
        target = self._get_closest(x, y, candidates)
        self.swap(win, target)

    @expose_command()
    def swap_right(self):
        """Swap current window with closest window to the up"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["y"] < y]
        target = self._get_closest(x, y, candidates)
        self.swap(win, target)

    @expose_command()
    def swap_main(self):
        """Swap current window to main pane"""
        if self.align == self._up:
            self.swap_right()
        elif self.align == self._down:
            self.swap_left()

    @expose_command()
    def left(self):
        """Focus on the closest window to the left of the current window"""
        self.previous()

    @expose_command()
    def right(self):
        """Focus on the closest window to the right of the current window"""
        self.next()

    @expose_command()
    def up(self):
        """Focus on the closest window above the current window"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["y"] < y]
        self.clients.current_client = self._get_closest(x, y, candidates)
        self.group.focus(self.clients.current_client)

    @expose_command()
    def down(self):
        """Focus on the closest window below the current window"""
        win = self.clients.current_client
        x, y = win.x, win.y
        candidates = [c for c in self.clients if c.info()["y"] > y]
        self.clients.current_client = self._get_closest(x, y, candidates)
        self.group.focus(self.clients.current_client)


class MonadThreeCol(MonadTall):
    """Emulate the behavior of XMonad's ThreeColumns layout.

    A layout similar to tall but with three columns. With an ultra wide display
    this layout can be used for a huge main window - ideally at the center of the
    screen - and up to six reasonable sized secondary windows.

    Main-Pane:

    A main pane that contains a single window takes up a vertical portion of
    the screen_rect based on the ratio setting. This ratio can be adjusted with
    the ``grow_main`` and ``shrink_main`` or, while the main pane is in
    focus, ``grow`` and ``shrink``. The main pane can also be centered.

    ::

        ---------------------------    ---------------------------
        |           |      |      |    |      |           |      |
        |           |      |      |    |      |           |      |
        |           |      |      |    |      |           |      |
        |           |      |      |    |      |           |      |
        |           |      |      |    |      |           |      |
        |           |      |      |    |      |           |      |
        ---------------------------    ---------------------------

    Secondary-panes:

    Occupying the rest of the screen_rect are one or more secondary panes.  The
    secondary panes will be divided into two columns and share the vertical space
    of each column. However they can be resized at will with the ``grow`` and
    ``shrink`` methods. The other secondary panes will adjust their sizes to
    smoothly fill all of the space.

    ::

        ---------------------------    ---------------------------
        |           |      |      |    |           |______|      |
        |           |______|      |    |           |      |      |
        |           |      |______|    |           |      |______|
        |           |______|      |    |           |      |      |
        |           |      |      |    |           |______|      |
        |           |      |      |    |           |      |      |
        ---------------------------    ---------------------------

    Panes can be moved with the ``shuffle_up`` and ``shuffle_down``
    methods. As mentioned the main pane is considered the top of the stack;
    moving up is counter-clockwise and moving down is clockwise. A secondary
    pane can also be promoted to the main pane with the ``swap_main``
    method.

    Normalizing/Resetting:

    To restore all secondary client windows to their default size ratios
    use the ``normalize`` method.

    To reset all client windows to their default sizes, including the primary
    window, use the ``reset`` method.

    Maximizing:

    To maximized a client window simply use the ``maximize`` on a focused
    client.
    """

    defaults = [
        ("main_centered", True, "Place the main pane at the center of the screen"),
        (
            "new_client_position",
            "top",
            "Place new windows: "
            " after_current - after the active window."
            " before_current - before the active window,"
            " top - at the top of the stack,"
            " bottom - at the bottom of the stack,",
        ),
    ]

    __column = namedtuple("__column", "name count start end")

    def __init__(self, **config):
        MonadTall.__init__(self, **config)
        self.add_defaults(MonadThreeCol.defaults)

    # mypy doesn't like the setter when the getter isn't present
    # see https://github.com/python/mypy/issues/5936
    @MonadTall.screen_rect.setter  # type: ignore[attr-defined]
    def screen_rect(self, value):
        # If the screen_rect size has change then we need to normalise secondary
        # windows so they're resized to fill the new space correctly
        if value != self._screen_rect:
            self.do_normalize = True
        self._screen_rect = value

    def _configure_specific(self, client, screen_rect, border_color, index):
        """Specific configuration for xmonad three columns."""
        if index == 0:
            self._configure_main(client)
        elif self._get_column(index - 1).name == "left":
            self._configure_left(client, index)
        else:
            self._configure_right(client, index)

    def _configure_main(self, client):
        """Configure the main client"""
        width = self._get_main_width()
        height = self.screen_rect.height
        left = self.screen_rect.x
        top = self.screen_rect.y

        if self.main_centered and len(self.clients) > 2:
            left += (self.screen_rect.width - width) // 2

        self._place_client(client, left, top, width, height)

    def _configure_left(self, client, index):
        """Configure the left column"""
        width = self._get_secondary_widths()[0]
        height = self._get_secondary_height(index)
        left = self.screen_rect.x
        top = self.screen_rect.y + self._get_relative_sizes_above(index)

        if not self.main_centered or len(self.clients) == 2:
            left += self._get_main_width()

        self._place_client(client, left, top, width, height)

    def _configure_right(self, client, index):
        """Configure the right column"""
        widths = self._get_secondary_widths()
        height = self._get_secondary_height(index)
        left = self.screen_rect.x + widths[0] + self._get_main_width()
        top = self.screen_rect.y + self._get_relative_sizes_above(index)

        self._place_client(client, left, top, widths[1], height)

    def _get_main_width(self):
        """Calculate the main client's width"""
        return int(self.screen_rect.width * self.ratio)

    def _get_secondary_widths(self):
        """Calculate secondary clients' widths"""
        width = self.screen_rect.width - self._get_main_width()
        if len(self.clients) == 2:
            return [width, 0]

        return self._split_integer(width, 2)

    def _get_secondary_height(self, index):
        """Return the height of the provided index"""
        return self.relative_sizes[index - 1]

    def _get_relative_sizes_above(self, index):
        """Return the sum of the heights of all clients above the provided index"""
        column = self._get_column(index - 1)
        return sum(self.relative_sizes[column.start : index - 1])

    def _place_client(self, client, left, top, width, height):
        """Place a client on the screen

        Will prevent double margins by applying east and south margins only
        when the client is the rightmost or the bottommost window.
        """

        # Create a temporary margin list for the client
        if isinstance(self.margin, int):
            margin = [self.margin] * 4
        else:
            # We need to copy this list otherwise we'd be modifying self.margin!
            margin = self.margin.copy()

        rightmost = left + width - self.screen_rect.x + margin[1] >= self.screen_rect.width
        bottommost = top + height - self.screen_rect.y + margin[2] >= self.screen_rect.height

        if not rightmost:
            margin[1] = 0
        if not bottommost:
            margin[2] = 0

        client.place(
            left,
            top,
            width - 2 * self.border_width,
            height - 2 * self.border_width,
            self.border_width,
            self.border_focus if client.has_focus else self.border_normal,
            margin=margin,
        )

    @expose_command()
    def normalize(self, redraw=True):
        """Evenly distribute screen-space among secondary clients"""
        if self.screen_rect is not None:
            self.relative_sizes = []

            height = self.screen_rect.height
            left, right = self._get_columns()

            if left.count > 0:
                self.relative_sizes += self._split_integer(height, left.count)
            if right.count > 0:
                self.relative_sizes += self._split_integer(height, right.count)

        if redraw:
            self.group.layout_all()
        self.do_normalize = False

    @expose_command()
    def swap_main(self):
        """Swap current window to main pane"""
        self.swap(self.clients.current_client, self.clients[0])

    def _maximize_secondary(self):
        """Maximize the focused secondary pane"""
        focused = self.focused - 1
        column = self._get_column(focused)
        if column.count == 1:
            return

        max_height = self.screen_rect.height - ((column.count - 1) * self.min_secondary_size)
        for i in range(column.start, column.end):
            self.relative_sizes[i] = max_height if i == focused else self.min_secondary_size

    def _grow_secondary(self, amt):
        """Grow the focused client in the secondary pane"""
        self._resize_secondary(amt)

    def _shrink_secondary(self, amt):
        """Shrink the focused client in the secondary pane"""
        self._resize_secondary(-amt)

    def _resize_secondary(self, amt):
        """Resize the focused secondary client

        If amt is positive, the client will grow. Conversely, if it's negative,
        the client will shrink. All other clients in the same column will get
        grown/shrunk so to accommodate the new height.
        """
        focused = self.focused - 1
        column = self._get_column(focused)

        if column.count == 1:
            return

        # Resizing is accomplished by doing the following:
        # - calculate how much each client in the column must shrink/grow
        #   so that the focused window can grow/shrink.
        # - iterate over all clients in the column and change their height
        #   (grow or shrink) as long as they can still be resized (both main
        #   and secondary windows).
        min_height = self.min_secondary_size
        idx = column.start
        step = amt // (column.count - 1)
        visited = 0
        while amt != 0:
            if idx != focused:
                focused_new_height = self.relative_sizes[focused] + step
                new_height = self.relative_sizes[idx] - step
                if focused_new_height >= min_height and new_height >= min_height:
                    self.relative_sizes[focused] += step
                    self.relative_sizes[idx] -= step
                    amt -= step
                    visited += 1

            idx += 1
            if idx == column.end:
                if visited == 0:
                    break

                idx = column.start
                visited = 0

        self.group.layout_all()

    def _get_column(self, index):
        """Get the column containing the provided index"""
        left, right = self._get_columns()
        return left if index < left.count else right

    def _get_columns(self):
        """Get all columns"""
        clients = len(self.clients) - 1
        clients = (clients // 2 + clients % 2, clients // 2)
        return (
            MonadThreeCol.__column(
                name="left",
                count=clients[0],
                start=0,
                end=clients[0],
            ),
            MonadThreeCol.__column(
                name="right",
                count=clients[1],
                start=clients[0],
                end=clients[0] + clients[1],
            ),
        )

    def info(self) -> dict[str, Any]:
        left, right = self._get_columns()
        d = MonadTall.info(self)
        d.update(
            secondary=dict(
                left=d["clients"][1 : left.end + 1] if left.count > 0 else [],
                right=d["clients"][right.start + 1 :] if right.count > 0 else [],
            )
        )
        return d

    @staticmethod
    def _split_integer(value, parts):
        """Divide an integer into equal parts and distribute the remainder"""
        result = [value // parts] * parts
        for i in range(value % parts):
            result[i] += 1
        return result



from __future__ import annotations

import os.path
import re
import sys
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING

from libqtile import configurable, hook, utils
from libqtile.bar import Bar
from libqtile.command.base import CommandObject, expose_command
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import Any, Literal

    from libqtile.backend import base
    from libqtile.backend.base.idle_notify import IdleAction
    from libqtile.bar import BarType
    from libqtile.command.base import ItemT
    from libqtile.core.manager import Qtile
    from libqtile.group import _Group
    from libqtile.layout.base import Layout
    from libqtile.lazy import LazyCall
    from libqtile.utils import ColorType


class Key:
    """
    Defines a keybinding.

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        ``"shift"``, ``"lock"``, ``"control"``, ``"mod1"``, ``"mod2"``, ``"mod3"``,
        ``"mod4"``, ``"mod5"``.
    key:
        A key specification, e.g. ``"a"``, ``"Tab"``, ``"Return"``, ``"space"``. Also accepts
        an integer value representing a keycode.
    commands:
        One or more :class:`LazyCall` objects to evaluate in sequence upon keypress. Multiple
        commands should be separated by commas.
    desc:
        Description to be added to the key binding. (Optional)
    swallow:
        Configures when we swallow the key binding. (Optional)
        Setting it to False will forward the key binding to the focused window after the commands have been executed.
    """

    def __init__(
        self,
        modifiers: list[str],
        key: str | int,
        *commands: LazyCall,
        desc: str = "",
        swallow: bool = True,
    ) -> None:
        self.modifiers = modifiers
        self.key = key
        self.commands = commands
        self.desc = desc
        self.swallow = swallow

    def __repr__(self) -> str:
        return f"<Key ({self.modifiers}, {self.key})>"


class KeyChord:
    """
    Define a key chord aka Vim-like mode.

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        ``"shift"``, ``"lock"``, ``"control"``, ``"mod1"``, ``"mod2"``, ``"mod3"``,
        ``"mod4"``, ``"mod5"``.
    key:
        A key specification, e.g. ``"a"``, ``"Tab"``, ``"Return"``, ``"space"``. Also accepts
        an integer value representing a keycode.
    submappings:
        A list of :class:`Key` or :class:`KeyChord` declarations to bind in this chord.
    mode:
        Boolean. Setting to ``True`` will result in the chord persisting until
        Escape is pressed. Setting to ``False`` (default) will exit the chord once
        the sequence has ended.
    name:
        A string to name the chord. The name will be displayed in the Chord
        widget.
    desc:
        A string to describe the chord. This attribute is not directly used by Qtile
        but users may want to access this when creating scripts to show configured
        keybindings.
    swallow:
        Configures when we swallow the key binding of the chord. (Optional)
        Setting it to False will forward the key binding to the focused window after the commands have been executed.
    """

    def __init__(
        self,
        modifiers: list[str],
        key: str | int,
        submappings: list[Key | KeyChord],
        mode: bool | str = False,
        name: str = "",
        desc: str = "",
        swallow: bool = True,
    ):
        self.modifiers = modifiers
        self.key = key

        submappings.append(Key([], "Escape"))
        self.submappings = submappings
        self.mode = mode
        self.name = name
        self.desc = desc

        if isinstance(mode, str):
            logger.warning(
                "The use of `mode` to set the KeyChord name is deprecated. "
                "Please use `name='%s'` instead. "
                "'mode' should be a boolean value to set whether the chord is persistent (True) or not.",
                mode,
            )
            self.name = mode
            self.mode = True
        self.swallow = swallow

    def __repr__(self) -> str:
        return f"<KeyChord ({self.modifiers}, {self.key})>"


class Mouse:
    def __init__(self, modifiers: list[str], button: str, *commands: LazyCall) -> None:
        self.modifiers = modifiers
        self.button = button
        self.commands = commands
        self.button_code = int(self.button.replace("Button", ""))
        self.modmask: int = 0


class Drag(Mouse):
    """
    Bind commands to a dragging action.

    On each motion event the bound commands are executed with two additional parameters
    specifying the x and y offset from the previous position.

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        ``"shift"``, ``"lock"``, ``"control"``, ``"mod1"``, ``"mod2"``, ``"mod3"``,
        ``"mod4"``, ``"mod5"``.
    button:
        The button used to start dragging e.g. ``"Button1"``.
    commands:
        A list :class:`LazyCall` objects to evaluate in sequence upon drag.
    start:
        A :class:`LazyCall` object to be evaluated when dragging begins. (Optional)
    warp_pointer:
        A :class:`bool` indicating if the pointer should be warped to the bottom right of the window
        at the start of dragging. (Default: `False`)

    """

    def __init__(
        self,
        modifiers: list[str],
        button: str,
        *commands: LazyCall,
        start: LazyCall | None = None,
        warp_pointer: bool = False,
    ) -> None:
        super().__init__(modifiers, button, *commands)
        self.start = start
        self.warp_pointer = warp_pointer

    def __repr__(self) -> str:
        return f"<Drag ({self.modifiers}, {self.button})>"


class Click(Mouse):
    """
    Bind commands to a clicking action.

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        ``"shift"``, ``"lock"``, ``"control"``, ``"mod1"``, ``"mod2"``, ``"mod3"``,
        ``"mod4"``, ``"mod5"``.
    button:
        The button used to click e.g. ``"Button1"``.
    commands:
        A list :class:`LazyCall` objects to evaluate in sequence upon click.

    """

    def __repr__(self) -> str:
        return f"<Click ({self.modifiers}, {self.button})>"


class EzConfig:
    """
    Helper class for defining key and button bindings in an Emacs-like format.

    Inspired by Xmonad's XMonad.Util.EZConfig.

    Splits an emacs keydef into modifiers and keys. For example:

          "m-s-a"     -> ['mod4', 'shift'], 'a'
          "a-<minus>" -> ['mod1'], 'minus'
          "C-<Tab>"   -> ['control'], 'Tab'

    """

    modifier_keys = {
        "M": "mod4",
        "A": "mod1",
        "S": "shift",
        "C": "control",
    }

    def parse(self, spec: str) -> tuple[list[str], str]:
        mods = []
        keys: list[str] = []

        for key in spec.split("-"):
            if not key:
                break
            if key in self.modifier_keys:
                if keys:
                    msg = "Modifiers must always come before key/btn: %s"
                    raise utils.QtileError(msg % spec)
                mods.append(self.modifier_keys[key])
                continue
            if len(key) == 1:
                keys.append(key)
                continue
            if len(key) > 3 and key[0] == "<" and key[-1] == ">":
                keys.append(key[1:-1])
                continue

        if not keys:
            msg = "Invalid key/btn specifier: %s"
            raise utils.QtileError(msg % spec)

        if len(keys) > 1:
            msg = f"Key chains are not supported: {spec}"
            raise utils.QtileError(msg)

        return mods, keys[0]


class EzKey(EzConfig, Key):
    """
    Defines a keybinding using the Emacs-like format.

    Parameters
    ==========
    keydef:
        The Emacs-like key specification, e.g. ``"M-S-a"``.
    commands:
        A list :class:`LazyCall` objects to evaluate in sequence upon keypress.
    desc:
        Description to be added to the key binding. (Optional)

    """

    def __init__(self, keydef: str, *commands: LazyCall, desc: str = "") -> None:
        modkeys, key = self.parse(keydef)
        super().__init__(modkeys, key, *commands, desc=desc)


class EzKeyChord(EzConfig, KeyChord):
    """
    Define a key chord using the Emacs-like format.

    Parameters
    ==========
    keydef:
        The Emacs-like key specification, e.g. ``"M-S-a"``.
    submappings:
        A list of :class:`Key` or :class:`KeyChord` declarations to bind in this chord.
    mode:
        Boolean. Setting to ``True`` will result in the chord persisting until
        Escape is pressed. Setting to ``False`` (default) will exit the chord once
        the sequence has ended.
    name:
        A string to name the chord. The name will be displayed in the Chord
        widget.
    desc:
        A string to describe the chord. This attribute is not directly used by Qtile
        but users may want to access this when creating scripts to show configured
        keybindings.

    """

    def __init__(
        self,
        keydef: str,
        submappings: list[Key | KeyChord],
        mode: bool | str = False,
        name: str = "",
        desc: str = "",
    ):
        modkeys, key = self.parse(keydef)
        super().__init__(modkeys, key, submappings, mode, name, desc)


class EzClick(EzConfig, Click):
    """
    Bind commands to a clicking action using the Emacs-like format.

    Parameters
    ==========
    btndef:
        The Emacs-like button specification, e.g. ``"M-1"``.
    commands:
        A list :class:`LazyCall` objects to evaluate in sequence upon drag.

    """

    def __init__(self, btndef: str, *commands: LazyCall) -> None:
        modkeys, button = self.parse(btndef)
        button = f"Button{button}"
        super().__init__(modkeys, button, *commands)


class EzDrag(EzConfig, Drag):
    """
    Bind commands to a dragging action using the Emacs-like format.

    Parameters
    ==========
    btndef:
        The Emacs-like button specification, e.g. ``"M-1"``.
    commands:
        A list :class:`LazyCall` objects to evaluate in sequence upon drag.
    start:
        A :class:`LazyCall` object to be evaluated when dragging begins. (Optional)

    """

    def __init__(self, btndef: str, *commands: LazyCall, start: LazyCall | None = None) -> None:
        modkeys, button = self.parse(btndef)
        button = f"Button{button}"
        super().__init__(modkeys, button, *commands, start=start)


@dataclass
class ScreenRect:
    x: int
    y: int
    width: int
    height: int

    def hsplit(self, columnwidth: int) -> tuple[ScreenRect, ScreenRect]:
        assert 0 < columnwidth < self.width
        return (
            self.__class__(self.x, self.y, columnwidth, self.height),
            self.__class__(self.x + columnwidth, self.y, self.width - columnwidth, self.height),
        )

    def vsplit(self, rowheight: int) -> tuple[ScreenRect, ScreenRect]:
        assert 0 < rowheight < self.height
        return (
            self.__class__(self.x, self.y, self.width, rowheight),
            self.__class__(self.x, self.y + rowheight, self.width, self.height - rowheight),
        )


@dataclass
class Output:
    port: str | None
    make: str | None
    model: str | None
    serial: str | None
    # Do not consider geometry when comparing outputs: we want to know if the
    # underlying physical hardware is the same, the geometry doesn't matter.
    rect: ScreenRect = field(compare=False)


class Screen(CommandObject):
    r"""
    A physical screen, and its associated paraphernalia.

    Define a screen with a given set of :class:`Bar`\s of a specific geometry. Also,
    ``x``, ``y``, ``width``, and ``height`` aren't specified usually unless you are
    using 'fake screens'.

    The ``background`` parameter, if given, should be a valid single colour. This will
    paint a solid background colour to the screen. Note, the setting is ignored if
    ``wallpaper`` is also set (see below).

    The ``wallpaper`` parameter, if given, should be a path to an image file. How this
    image is painted to the screen is specified by the ``wallpaper_mode`` parameter. By
    default, the image will be placed at the screens origin and retain its own
    dimensions. If the mode is ``"fill"``, the image will be centred on the screen and
    resized to fill it. If the mode is ``"stretch"``, the image is stretched to fit all
    of it into the screen. If the mode is ``"center"``, the image is centered
    on the screen. The ``"background"`` painting is done before the image is
    drawn, so you can set the background color for an image by setting
    background here.

    The ``x11_drag_polling_rate`` parameter specifies the rate for drag events
    in the X11 backend. By default this is set to None, indicating no limit.
    Because in the X11 backend we already handle motion notify events later,
    the performance should already be okay. However, to limit these events
    further you can use this variable and e.g. set it to your monitor refresh
    rate. 60 would mean that we handle a drag event 60 times per second.
    """

    group: _Group
    index: int
    # This is populated in manager.py's _process_screens()
    output: Output

    def __init__(
        self,
        top: BarType | None = None,
        bottom: BarType | None = None,
        left: BarType | None = None,
        right: BarType | None = None,
        background: ColorType | None = None,
        wallpaper: str | None = None,
        wallpaper_mode: str | None = None,
        x11_drag_polling_rate: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        self.top = top
        self.bottom = bottom
        self.left = left
        self.right = right
        self.background = background
        self.wallpaper = wallpaper
        self.wallpaper_mode = wallpaper_mode
        self.x11_drag_polling_rate = x11_drag_polling_rate
        self.qtile: Qtile | None = None
        # x position of upper left corner can be > 0
        # if one screen is "right" of the other
        self.x = x if x is not None else 0
        self.y = y if y is not None else 0
        self.width = width if width is not None else 0
        self.height = height if height is not None else 0
        self.previous_group: _Group | None = None
        self.output = Output(None, None, None, None, ScreenRect(0, 0, 0, 0))

    def __eq__(self, other: object) -> bool:
        # When we trigger a reconfigure_screens(), _process_screens()
        # re-creates the Screen() objects. Since we only replace qtile.screens,
        # groups, widgets, etc. which have saved a reference to the Screen
        # object will cause comparisons to fail even if a Screen object after
        # reconfiguration represents the same geometry area, which is generally
        # not what people expect.
        #
        # For example, this happens here:
        # https://github.com/qtile/qtile/blob/42e03cdb6bbb6d88f0fd58927c6ec6258f625961/libqtile/group.py#L167-L168
        # which does:
        #
        #     if self.current_window and self.screen == self.qtile.current_screen:
        #         self.current_window.focus(warp)
        #
        # This will fail, because the group's object is stale, and we will not
        # focus things correctly. This is one example, but there are several
        # other object trees that save copies of the current screen.
        if not isinstance(other, Screen):
            return False

        # Use output matching if we have identifying info (port, make, model, or serial).
        # This allows us to reason correctly about focus when monitor geometry changes.
        if (
            self.output.port is not None
            or self.output.make is not None
            or self.output.model is not None
            or self.output.serial is not None
        ):
            return self.output == other.output

        # Fall back to geometry comparison when output info isn't available
        # (e.g., in xephyr, which doesn't export any edid info)
        return (
            other.x == self.x
            and other.y == self.y
            and other.width == self.width
            and other.height == self.height
        )

    def __hash__(self) -> int:
        return hash((self.x, self.y, self.width, self.height))

    def _configure(
        self,
        qtile: Qtile,
        index: int,
        x: int,
        y: int,
        width: int,
        height: int,
        group: _Group,
        reconfigure_gaps: bool = False,
    ) -> None:
        self.qtile = qtile
        self.index = index
        self.x = x
        self.y = y
        self.width = width
        self.height = height

        for gap in self.gaps:
            try:
                gap._configure(qtile, self, reconfigure=reconfigure_gaps)
            except Exception:
                logger.exception(f"Error configuring {gap.position} gap/bar.")
                # Call finalize to prevent future execution of the _actual_draw method
                self.finalize_gap(gap.position)

        self.set_group(group)

        if self.background is not None:
            self.qtile.fill_screen(self, self.background)
        if self.wallpaper is not None:
            self.wallpaper = os.path.expanduser(self.wallpaper)
            self.paint(self.wallpaper, self.wallpaper_mode)

    def paint(self, path: str, mode: str | None = None) -> None:
        if self.qtile:
            self.qtile.paint_screen(self, path, mode)

    @property
    def gaps(self) -> Iterable[BarType]:
        return (i for i in [self.top, self.bottom, self.left, self.right] if i)

    def finalize_gap(self, position: str) -> None:
        gap = getattr(self, position, None)
        if gap is not None:
            gap.finalize()

    def finalize_gaps(self) -> None:
        for position in ["top", "bottom", "left", "right"]:
            self.finalize_gap(position)

    @property
    def dx(self) -> int:
        if self.left and getattr(self.left, "reserve", True):
            return self.x + self.left.fullsize
        return self.x

    @property
    def dy(self) -> int:
        if self.top and getattr(self.top, "reserve", True):
            return self.y + self.top.fullsize
        return self.y

    @property
    def dwidth(self) -> int:
        val = self.width
        if self.left and getattr(self.left, "reserve", True):
            val -= self.left.fullsize
        if self.right and getattr(self.right, "reserve", True):
            val -= self.right.fullsize
        return val

    @property
    def dheight(self) -> int:
        val = self.height
        if self.top and getattr(self.top, "reserve", True):
            val -= self.top.fullsize
        if self.bottom and getattr(self.bottom, "reserve", True):
            val -= self.bottom.fullsize
        return val

    def get_rect(self) -> ScreenRect:
        return ScreenRect(self.dx, self.dy, self.dwidth, self.dheight)

    def set_group(
        self, new_group: _Group | None, save_prev: bool = True, warp: bool = True
    ) -> None:
        """Put group on this screen"""
        if new_group is None:
            return

        if new_group.screen == self:
            return

        if save_prev and new_group is not self.group:
            # new_group can be self.group only if the screen is getting configured for
            # the first time
            self.previous_group = self.group

        if new_group.screen:
            # g1 <-> s1 (self)
            # g2 (new_group) <-> s2 to
            # g1 <-> s2
            # g2 <-> s1
            g1 = self.group
            s1 = self
            g2 = new_group
            s2 = new_group.screen

            s2.group = g1
            g1.set_screen(s2, warp)
            s1.group = g2
            g2.set_screen(s1, warp)
        else:
            assert self.qtile is not None
            old_group = self.group
            self.group = new_group
            with self.qtile.core.masked():
                # display clients of the new group and then hide from old group
                # to remove the screen flickering
                new_group.set_screen(self, warp)

                # Can be the same group only if the screen just got configured for the
                # first time - see `Qtile._process_screens`.
                if old_group is not new_group:
                    old_group.set_screen(None, warp)

        hook.fire("setgroup")
        hook.fire("focus_change")
        hook.fire("layout_change", self.group.layouts[self.group.current_layout], self.group)

    def _toggle_group(self, group: _Group | None = None, warp: bool = True) -> None:
        """Switch to the selected group or to the previously active one"""
        if group in (self.group, None) and self.previous_group:
            group = self.previous_group
        self.set_group(group, warp=warp)

    def _items(self, name: str) -> ItemT:
        if name == "layout" and self.group is not None:
            return True, list(range(len(self.group.layouts)))
        elif name == "window" and self.group is not None:
            return True, [i.wid for i in self.group.windows]
        elif name == "bar":
            return False, [x.position for x in self.gaps if isinstance(x, Bar)]
        elif name == "widget":
            bars = (g for g in self.gaps if isinstance(g, Bar))
            return False, [w.name for b in bars for w in b.widgets]
        elif name == "group":
            return True, [self.group.name]
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "layout":
            if sel is None:
                return self.group.layout
            else:
                assert isinstance(sel, int)
                return utils.lget(self.group.layouts, sel)
        elif name == "window":
            if sel is None:
                return self.group.current_window
            else:
                for i in self.group.windows:
                    if i.wid == sel:
                        return i
        elif name == "bar":
            assert isinstance(sel, str)
            bar = getattr(self, sel)
            if isinstance(bar, Bar):
                return bar

        elif name == "widget":
            for gap in self.gaps:
                if not isinstance(gap, Bar):
                    continue
                for widget in gap.widgets:
                    if widget.name == sel:
                        return widget
        elif name == "group":
            if sel is None:
                return self.group
            else:
                return self.group if sel == self.group.name else None
        return None

    @expose_command
    def resize(
        self,
        x: int | None = None,
        y: int | None = None,
        w: int | None = None,
        h: int | None = None,
    ) -> None:
        assert self.qtile is not None
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if w is None:
            w = self.width
        if h is None:
            h = self.height
        self._configure(self.qtile, self.index, x, y, w, h, self.group)
        for bar in [self.top, self.bottom, self.left, self.right]:
            if bar:
                bar.draw()
        self.group.layout_all()

    @expose_command()
    def info(self) -> dict[str, Any]:
        """Returns a dictionary of info for this screen."""
        return dict(
            index=self.index,
            width=self.width,
            height=self.height,
            x=self.x,
            y=self.y,
            **asdict(self.output),
        )

    @expose_command()
    def next_group(
        self, skip_empty: bool = False, skip_managed: bool = False, warp: bool = True
    ) -> None:
        """Switch to the next group"""
        group = self.group.get_next_group(skip_empty, skip_managed)
        self.set_group(group, warp=warp)

    @expose_command()
    def prev_group(
        self, skip_empty: bool = False, skip_managed: bool = False, warp: bool = True
    ) -> None:
        """Switch to the previous group"""
        group = self.group.get_previous_group(skip_empty, skip_managed)
        self.set_group(group, warp=warp)

    @expose_command()
    def toggle_group(self, group_name: str | None = None, warp: bool = True) -> None:
        """Switch to the selected group or to the previously active one"""
        assert self.qtile is not None
        group = self.qtile.groups_map.get(group_name if group_name else "")
        self._toggle_group(group, warp=warp)

    @expose_command()
    def set_wallpaper(self, path: str, mode: str | None = None) -> None:
        """Set the wallpaper to the given file."""
        self.paint(path, mode)


class Group:
    """
    Represents a "dynamic" group

    These groups can spawn apps, only allow certain Matched windows to be on them, hide
    when they're not in use, etc. Groups are identified by their name.

    Parameters
    ==========
    name:
        The name of this group.
    matches:
        List of :class:`Match` objects whose matched windows will be assigned to this
        group.
    exclusive:
        When other apps are started in this group, should we allow them here or not?
    spawn:
        This will be executed (via ``qtile.spawn()``) when the group is created. You can pass either a
        program name or a list of programs to ``exec()``.
    layout:
        The name of default layout for this group (e.g. ``"max"``). This is the name
        specified for a particular layout in ``config.py`` or if not defined it defaults
        in general to the class name in all lower case.
    layouts:
        The group layouts list overriding global layouts. Use this to define a separate
        list of layouts for this particular group.
    persist:
        Should this group stay alive when it has no member windows?
    init:
        Should this group be alive when Qtile starts?
    layout_opts:
        Options to pass to a layout.
    screen_affinity:
        Make a dynamic group prefer to start on a specific screen.
    position:
        The position of this group.
    label:
        The display name of the group. Use this to define a display name other than name
        of the group. If set to ``None``, the display name is set to the name.

    """

    def __init__(
        self,
        name: str,
        matches: list[Match] | None = None,
        exclusive: bool = False,
        spawn: str | list[str] | None = None,
        layout: str | None = None,
        layouts: list[Layout] | None = None,
        persist: bool = True,
        init: bool = True,
        layout_opts: dict[str, Any] | None = None,
        screen_affinity: int | None = None,
        position: int = sys.maxsize,
        label: str | None = None,
    ) -> None:
        self.name = name
        self.label = label
        self.exclusive = exclusive
        self.spawn = spawn
        self.layout = layout
        self.layouts = layouts or []
        self.persist = persist
        self.init = init
        self.matches = matches or []
        self.layout_opts = layout_opts or {}

        self.screen_affinity = screen_affinity
        self.position = position

    def __repr__(self) -> str:
        attrs = utils.describe_attributes(
            self,
            [
                "exclusive",
                "spawn",
                "layout",
                "layouts",
                "persist",
                "init",
                "matches",
                "layout_opts",
                "screen_affinity",
            ],
        )
        return f"<config.Group {self.name!r} ({attrs})>"


class ScratchPad(Group):
    """
    Represents a "ScratchPad" group

    ScratchPad adds a (by default) invisible group to Qtile. That group is used as a
    place for currently not visible windows spawned by a :class:`DropDown`
    configuration.

    Parameters
    ==========
    name:
        The name of this group.
    dropdowns:
        :class:`DropDown` s available on the scratchpad.
    position:
        The position of this group.
    label:
        The display name of the :class:`ScratchPad` group. Defaults to the empty string
        such that the group is hidden in :class:`~libqtile.widget.GroupBox` widget.
    single:
        If ``True``, only one of the dropdowns will be visible at a time.

    """

    def __init__(
        self,
        name: str,
        dropdowns: list[DropDown] | None = None,
        position: int = sys.maxsize,
        label: str = "",
        single: bool = False,
    ) -> None:
        Group.__init__(
            self,
            name,
            layout="floating",
            init=False,
            position=position,
            label=label,
        )
        self.dropdowns = dropdowns if dropdowns is not None else []
        self.single = single

    def __repr__(self) -> str:
        return "<config.ScratchPad {!r} ({})>".format(
            self.name,
            ", ".join(dd.name for dd in self.dropdowns),
        )


def convert_deprecated_list(vals: list[str], name: str) -> re.Pattern:
    # make a regex with OR on word boundaries
    regex_input = r"^({})$".format("|".join(map(re.escape, vals)))
    logger.warning(
        "Your Match with the %s property is using lists which are deprecated, replace Match(%s=%s) with Match(%s=re.compile(r\"%s\")) after importing the 're' module",
        name,
        name,
        vals,
        name,
        regex_input,
    )
    return re.compile(regex_input)


class _Match:
    """Base class to implement bitwise logic methods for Match objects."""

    def compare(self, client: base.Window) -> bool:
        return True

    def __invert__(self) -> InvertMatch:
        return InvertMatch(self)

    def __and__(self, other: _Match) -> MatchAll:
        if not isinstance(other, _Match):
            raise TypeError

        return MatchAll(self, other)

    def __or__(self, other: _Match) -> MatchAny:
        if not isinstance(other, _Match):
            raise TypeError

        return MatchAny(self, other)

    def __xor__(self, other: _Match) -> MatchOnlyOne:
        if not isinstance(other, _Match):
            raise TypeError

        return MatchOnlyOne(self, other)


class InvertMatch(_Match):
    """Wrapper to invert the result of the comparison."""

    def __init__(self, match: _Match):
        self.match = match

    def compare(self, client: base.Window) -> bool:
        return not self.match.compare(client)

    def __repr__(self) -> str:
        return f"<InvertMatch({self.match!r})>"


class MatchAll(_Match):
    """Wrapper to check if all comparisons return True."""

    def __init__(self, *matches: _Match):
        self.matches = matches

    def compare(self, client: base.Window) -> bool:
        return all(m.compare(client) for m in self.matches)

    def __repr__(self) -> str:
        return f"<MatchAll({self.matches!r})>"


class MatchAny(MatchAll):
    """Wrapper to check if at least one of the comparisons returns True."""

    def compare(self, client: base.Window) -> bool:
        return any(m.compare(client) for m in self.matches)

    def __repr__(self) -> str:
        return f"<MatchAny({self.matches!r})>"


class MatchOnlyOne(_Match):
    """Wrapper to check if only one of the two comparisons returns True."""

    def __init__(self, match1: _Match, match2: _Match):
        self.match1 = match1
        self.match2 = match2

    def compare(self, client: base.Window) -> bool:
        return self.match1.compare(client) != self.match2.compare(client)

    def __repr__(self) -> str:
        return f"<MatchOnlyOne({self.match1!r}, {self.match2!r})>"


class Match(_Match):
    """
    Window properties to compare (match) with a window.

    The properties will be compared to a :class:`~libqtile.base.Window` to determine if
    its properties *match*. It can match by title, wm_class, role, wm_type,
    wm_instance_class, net_wm_pid, or wid. Additionally, a function may be
    passed, which takes in the :class:`~libqtile.base.Window` to be compared
    against and returns a boolean.

    For some properties, :class:`Match` supports both regular expression objects (i.e.
    the result of ``re.compile()``) or strings (match as an exact string). If a
    window matches all specified values, it is considered a match.

    Parameters
    ==========
    title:
        Match against the WM_NAME atom (X11) or title (Wayland).
    wm_class:
        Match against any value in the whole WM_CLASS atom (X11) or app ID (Wayland).
    role:
        Match against the WM_ROLE atom (X11 only).
    wm_type:
        Match against the WM_TYPE atom (X11 only).
    wm_instance_class:
        Match against the first string in WM_CLASS atom (X11) or app ID (Wayland).
    net_wm_pid:
        Match against the _NET_WM_PID atom (X11) or PID (Wayland).
    func:
        Delegate the match to the given function, which receives the tested client as an
        argument and must return ``True`` if it matches, ``False`` otherwise.
    wid:
        Match against the window ID. This is a unique ID given to each window.

    """

    def __init__(
        self,
        title: str | re.Pattern | None = None,
        wm_class: str | re.Pattern | None = None,
        role: str | re.Pattern | None = None,
        wm_type: str | re.Pattern | None = None,
        wm_instance_class: str | re.Pattern | None = None,
        net_wm_pid: int | None = None,
        func: Callable[[base.Window], bool] | None = None,
        wid: int | None = None,
    ) -> None:
        self._rules: dict[str, Any] = {}

        if title is not None:
            if isinstance(title, list):  # type: ignore
                title = convert_deprecated_list(title, "title")  # type: ignore
            self._rules["title"] = title
        if wm_class is not None:
            if isinstance(wm_class, list):  # type: ignore
                wm_class = convert_deprecated_list(wm_class, "wm_class")  # type: ignore
            self._rules["wm_class"] = wm_class
        if wm_instance_class is not None:
            if isinstance(wm_instance_class, list):  # type: ignore
                wm_instance_class = convert_deprecated_list(  # type: ignore
                    wm_instance_class, "wm_instance_class"
                )
            self._rules["wm_instance_class"] = wm_instance_class
        if wid is not None:
            self._rules["wid"] = wid
        if net_wm_pid is not None:
            try:
                self._rules["net_wm_pid"] = int(net_wm_pid)
            except ValueError:
                error = f'Invalid rule for net_wm_pid: "{str(net_wm_pid)}" only int allowed'
                raise utils.QtileError(error)
        if func is not None:
            self._rules["func"] = func

        if role is not None:
            if isinstance(role, list):  # type: ignore
                role = convert_deprecated_list(role, "role")  # type: ignore
            self._rules["role"] = role
        if wm_type is not None:
            if isinstance(wm_type, list):  # type: ignore
                wm_type = convert_deprecated_list(wm_type, "wm_type")  # type: ignore
            self._rules["wm_type"] = wm_type

    @staticmethod
    def _get_property_predicate(name: str, value: Any) -> Callable[..., bool]:
        if name == "net_wm_pid" or name == "wid":
            return lambda other: other == value
        elif name == "wm_class":

            def predicate(other) -> bool:
                match = getattr(other, "match", lambda v: v == other)
                return value and any(match(v) for v in value)

            return predicate
        else:

            def predicate(other) -> bool:
                match = getattr(other, "match", lambda v: v == other)
                return match(value)

            return predicate

    def compare(self, client: base.Window) -> bool:
        value: Any
        for property_name, rule_value in self._rules.items():
            if property_name == "title":
                value = client.name
            elif "class" in property_name:
                wm_class = client.get_wm_class()
                if not wm_class:
                    return False
                if property_name == "wm_instance_class":
                    value = wm_class[0]
                else:
                    value = wm_class
            elif property_name == "role":
                value = client.get_wm_role()
            elif property_name == "func":
                return rule_value(client)
            elif property_name == "net_wm_pid":
                value = client.get_pid()
            elif property_name == "wid":
                value = client.wid
            else:
                value = client.get_wm_type()

            # Some of the window.get_...() functions can return None
            if value is None:
                return False

            match = self._get_property_predicate(property_name, value)
            if not match(rule_value):
                return False

        if not self._rules:
            return False
        return True

    def map(self, callback: Callable[[base.Window], Any], clients: list[base.Window]) -> None:
        """Apply callback to each client that matches this Match"""
        for c in clients:
            if self.compare(c):
                callback(c)

    def __repr__(self) -> str:
        return f"<Match {self._rules}>"


class Rule:
    """
    How to act on a match.

    A :class:`Rule` contains a list of :class:`Match` objects, and a specification about
    what to do when any of them is matched.

    Parameters
    ==========
    match:
        :class:`Match` object or a list of such associated with this rule.
    float:
        Should we auto float this window?
    intrusive:
        Should we override the group's exclusive setting?
    break_on_match:
        Should we stop applying rules if this rule is matched?
    one_time:
        Should we delete rule once if it has been matched?

    """

    def __init__(
        self,
        match: _Match | list[_Match],
        group: _Group | None = None,
        float: bool = False,
        intrusive: bool = False,
        break_on_match: bool = True,
        one_time: bool = False,
    ) -> None:
        if isinstance(match, _Match):
            self.matchlist = [match]
        else:
            self.matchlist = match
        self.group = group
        self.float = float
        self.intrusive = intrusive
        self.break_on_match = break_on_match
        self.one_time = one_time

    def matches(self, w: base.Window) -> bool:
        return any(w.match(m) for m in self.matchlist)

    def __repr__(self) -> str:
        actions = utils.describe_attributes(
            self, ["group", "float", "intrusive", "break_on_match"]
        )
        return f"<Rule match={self.matchlist!r} actions=({actions})>"


class IdleTimer:
    """
    Creates a timer that will trigger an action when the system has been idle
    for a specified amount of time. An action can also be triggered when user input
    has been detected (NB this will only fire if the idle time is greater or equal to
    the specified timeout).

    IdleNotifier takes the following arguments:
      - ``timeout``: (int) timeout in seconds
      - ``action``: (callable or ``LazyCall``) the action to run when the timeout period is met (optional).
      - ``resume``: (callable or ``LazyCall``) the action run when user input is detected (optional).
      - ``respect_inhibitor``: (boolean) whether the action should be fired if there is an active idle inhibitor
                               (default ``True``).

    ``action`` and ``resume`` can also take coroutines so that the actions are run asynchronously.

    At least one of ``action`` and ``resume`` must be set.
    """

    def __init__(
        self,
        timeout: int,
        action: IdleAction = None,
        resume: IdleAction = None,
        respect_inhibitor: bool = True,
    ) -> None:
        if not (action or resume):
            raise ValueError("You must set one of 'action' or 'resume'.")

        if not isinstance(timeout, int) or timeout < 0:
            raise ValueError(f"Invalid idle timeout specified: {timeout}.")

        self.timeout = timeout
        self.action = action
        self.resume = resume
        self.respect_inhibitor = respect_inhibitor
        self.fired = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, IdleTimer):
            return False
        return self.timeout == other.timeout

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, IdleTimer):
            return False
        return self.timeout < other.timeout


class IdleInhibitor:
    """
    Create rules for when the compositor should not go into an idle state.

    IdleInhibitor take two arguments:

      -  match: a ``Match`` object to define which windows the rule should apply to. If unset, it will apply to all windows.
                Note: qtile evaluates whether a rule matches a window once, when the window is first created.
      -  when: one of the following strings:

        - "focus" (default): Inhibitor is active when the matching window is the currently focused window
        - "fullscreen": Inhibitor is active when the matching window is fullscreen
        - "visible": Inhibitor is active when the matching window is visible on any screen
                     (still applies if window is completely covered by a floating window)
        - "open": Inhibitor is active when the matching window is open (even if hidden)

    """

    def __init__(
        self,
        match: _Match | None = None,
        when: Literal["focus" | "fullscreen" | "visible" | "open"] = "open",
    ):
        self.match = match
        self.when = when


class DropDown(configurable.Configurable):
    """
    Configure a specified command and its associated window for the :class:`ScratchPad`.
    That window can be shown and hidden using a configurable keystroke or any other
    scripted trigger.
    """

    defaults = (
        (
            "x",
            0.1,
            "X position of window as fraction of current screen width. "
            "0 is the left most position.",
        ),
        (
            "y",
            0.0,
            "Y position of window as fraction of current screen height. "
            "0 is the top most position. To show the window at bottom, "
            "you have to configure a value < 1 and an appropriate height.",
        ),
        ("width", 0.8, "Width of window as fraction of current screen width"),
        ("height", 0.35, "Height of window as fraction of current screen."),
        ("opacity", 0.9, "Opacity of window as fraction. One is opaque."),
        (
            "on_focus_lost_hide",
            True,
            "Shall the window be hidden if focus is lost? If so, the :class:`DropDown` "
            "is hidden if window focus or the group is changed.",
        ),
        (
            "warp_pointer",
            True,
            "Shall pointer warp to center of window on activation? "
            "This only has effect if any of the ``on_focus_lost_xxx`` options are "
            "``True``",
        ),
        (
            "match",
            None,
            "Use a :class:`Match` to identify the spawned window and move it to the "
            "scratchpad, instead of relying on the window's PID. This works around "
            "some programs that may not be caught by the window's PID if it does "
            "not match the PID of the spawned process.",
        ),
    )

    def __init__(self, name: str, cmd: str, **config: Any) -> None:
        """
        Initialize :class:`DropDown` window wrapper.

        Define a command to spawn a process for the first time the class:`DropDown` is
        shown.

        Parameters
        ==========
        name:
            The name of the dropdown.
        cmd:
            Command to spawn a window to be captured by the dropdown.
        """
        configurable.Configurable.__init__(self, **config)
        self.name = name
        self.command = cmd
        self.add_defaults(self.defaults)

    def info(self) -> dict[str, Any]:
        return dict(
            name=self.name,
            command=self.command,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            opacity=self.opacity,
            on_focus_lost_hide=self.on_focus_lost_hide,
            warp_pointer=self.warp_pointer,
        )



import argparse
import logging
import sys
from pathlib import Path

from libqtile.log_utils import get_default_log, init_log
from libqtile.scripts import (
    check,
    cmd_obj,
    launch,
    migrate,
    repl,
    run_cmd,
    shell,
    start,
    top,
    x11_identify_output,
)
from libqtile.utils import VERSION


def main():
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-l",
        "--log-level",
        default="WARNING",
        dest="log_level",
        type=str.upper,
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        help="Set qtile log level",
    )
    parent_parser.add_argument(
        "-p",
        "--log-path",
        default=get_default_log(),
        dest="log_path",
        type=Path,
        help="Set alternative qtile log path",
    )
    main_parser = argparse.ArgumentParser(
        prog="qtile",
        description="A full-featured tiling window manager for X11 and Wayland.",
    )
    main_parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=VERSION,
    )

    subparsers = main_parser.add_subparsers()
    start.add_subcommand(subparsers, [parent_parser])
    shell.add_subcommand(subparsers, [parent_parser])
    top.add_subcommand(subparsers, [parent_parser])
    run_cmd.add_subcommand(subparsers, [parent_parser])
    cmd_obj.add_subcommand(subparsers, [parent_parser])
    check.add_subcommand(subparsers, [parent_parser])
    migrate.add_subcommand(subparsers, [parent_parser])
    launch.add_subcommand(subparsers, [parent_parser])
    repl.add_subcommand(subparsers, [parent_parser])
    x11_identify_output.add_subcommand(subparsers, [parent_parser])

    # `qtile help` should print help
    def print_help(options):
        main_parser.print_help()

    help_ = subparsers.add_parser(
        "help", help="Print help message and exit.", parents=[parent_parser]
    )
    help_.set_defaults(func=print_help)

    options = main_parser.parse_args()
    if func := getattr(options, "func", None):
        log_level = getattr(logging, options.log_level)
        init_log(log_level, log_path=options.log_path)
        func(options)
    else:
        main_parser.print_usage()
        print("")
        print("Did you mean:")
        print(" ".join(sys.argv + ["start"]))
        sys.exit(1)


if __name__ == "__main__":
    main()



import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Change,
    Check,
    MigrationTransformer,
    NoChange,
    _QtileMigrator,
    add_migration,
)


class StocktickerArgsTransformer(MigrationTransformer):
    @m.call_if_inside(
        m.Call(func=m.Name("StockTicker")) | m.Call(func=m.Attribute(attr=m.Name("StockTicker")))
    )
    @m.leave(m.Arg(keyword=m.Name("function")))
    def update_stockticker_args(self, original_node, updated_node) -> cst.Arg:
        """Changes 'function' kwarg to 'mode' and 'func' kwargs."""
        self.lint(original_node, "The 'function' keyword argument should be renamed 'func'.")
        return updated_node.with_changes(keyword=cst.Name("func"))


class StocktickerArgs(_QtileMigrator):
    ID = "UpdateStocktickerArgs"

    SUMMARY = "Updates ``StockTicker`` argument signature."

    HELP = """
    The ``StockTicker`` widget had a keyword argument called ``function``. This needs to be
    renamed to ``func`` to prevent clashes with the ``function()`` method of ``CommandObject``.

    For example:

    .. code:: python

        widget.StockTicker(function="TIME_SERIES_INTRADAY")

    should be changed to:

    .. code::

        widget.StockTicker(func="TIME_SERIES_INTRADAY")

    """

    AFTER_VERSION = "0.22.1"

    TESTS = [
        Change(
            """StockTicker(function="TIME_SERIES_INTRADAY")""",
            """StockTicker(func="TIME_SERIES_INTRADAY")""",
        ),
        Change(
            """widget.StockTicker(function="TIME_SERIES_INTRADAY")""",
            """widget.StockTicker(func="TIME_SERIES_INTRADAY")""",
        ),
        Change(
            """libqtile.widget.StockTicker(function="TIME_SERIES_INTRADAY")""",
            """libqtile.widget.StockTicker(func="TIME_SERIES_INTRADAY")""",
        ),
        NoChange("""StockTicker(func="TIME_SERIES_INTRADAY")"""),
        NoChange("""widget.StockTicker(func="TIME_SERIES_INTRADAY")"""),
        NoChange("""libqtile.widget.StockTicker(func="TIME_SERIES_INTRADAY")"""),
        Check(
            """
            import libqtile
            from libqtile import bar, widget
            from libqtile.widget import StockTicker

            bar.Bar(
                [
                    StockTicker(function="TIME_SERIES_INTRADAY"),
                    widget.StockTicker(function="TIME_SERIES_INTRADAY"),
                    libqtile.widget.StockTicker(function="TIME_SERIES_INTRADAY")
                ],
                20
            )
            """,
            """
            import libqtile
            from libqtile import bar, widget
            from libqtile.widget import StockTicker

            bar.Bar(
                [
                    StockTicker(func="TIME_SERIES_INTRADAY"),
                    widget.StockTicker(func="TIME_SERIES_INTRADAY"),
                    libqtile.widget.StockTicker(func="TIME_SERIES_INTRADAY")
                ],
                20
            )
            """,
        ),
    ]

    visitor = StocktickerArgsTransformer


add_migration(StocktickerArgs)



import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class BluetoothArgsTransformer(MigrationTransformer):
    @m.call_if_inside(
        m.Call(func=m.Name("Bluetooth")) | m.Call(func=m.Attribute(attr=m.Name("Bluetooth")))
    )
    @m.leave(m.Arg(keyword=m.Name("hci")))
    def update_bluetooth_args(self, original_node, updated_node) -> cst.Arg:
        """Changes positional  argumentto 'widgets' kwargs."""
        self.lint(
            original_node,
            "The 'hci' argument is deprecated and should be replaced with 'device'.",
        )
        return updated_node.with_changes(keyword=cst.Name("device"))


class BluetoothArgs(_QtileMigrator):
    ID = "UpdateBluetoothArgs"

    SUMMARY = "Updates ``Bluetooth`` argument signature."

    HELP = """
    The ``Bluetooth`` widget previously accepted a ``hci`` keyword argument. This has
    been deprecated following a major overhaul of the widget and should be replaced with
    a keyword argument named ``device``.

    For example:

    .. code:: python

        widget.Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX")

    should be changed to:

    .. code::

        widget.Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX")

    """

    AFTER_VERSION = "0.23.0"

    TESTS = [
        Check(
            """
            from libqtile import bar, widget
            from libqtile.widget import Bluetooth

            bar.Bar(
                [
                    Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX"),
                    widget.Bluetooth(hci="/dev_XX_XX_XX_XX_XX_XX"),
                ],
                20,
            )
            """,
            """
            from libqtile import bar, widget
            from libqtile.widget import Bluetooth

            bar.Bar(
                [
                    Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX"),
                    widget.Bluetooth(device="/dev_XX_XX_XX_XX_XX_XX"),
                ],
                20,
            )
            """,
        ),
    ]

    visitor = BluetoothArgsTransformer


add_migration(BluetoothArgs)



from __future__ import annotations

import fcntl
import os
from typing import TYPE_CHECKING

from libqtile import hook
from libqtile.log_utils import logger
from libqtile.utils import create_task

if TYPE_CHECKING:
    from typing import Any

try:
    from dbus_fast.aio import MessageBus
    from dbus_fast.constants import BusType
    from dbus_fast.errors import DBusError

    has_dbus = True
except ImportError:
    has_dbus = False

LOGIND_SERVICE = "org.freedesktop.login1"
LOGIND_INTERFACE = LOGIND_SERVICE + ".Manager"
LOGIND_PATH = "/org/freedesktop/login1"


class Inhibitor:
    """
    Class definition to access systemd's login1 service on dbus.

    Listens for `PrepareForSleep` signals and fires appropriate hooks
    when the signal is received.

    Can also set a sleep inhibitor which will be run after the "suspend"
    hook has been fired. This helps hooked functions to complete before
    the system goes to sleep. However, the inhibitor is set to only delay
    sleep, not block it.
    """

    def __init__(self) -> None:
        self.bus: MessageBus | None = None
        self.sleep = False
        self.resume = False
        self.fd: int = -1
        if TYPE_CHECKING:
            self.login: Any

    def want_sleep(self) -> None:
        """
        Convenience method to set flag to show we want to know when the
        system is going down for sleep.
        """
        if not has_dbus:
            logger.warning("dbus-fast must be installed to listen to sleep signals")
        self.sleep = True

    def want_resume(self) -> None:
        """
        Convenience method to set flag to show we want to know when the
        system is waking from sleep.
        """
        if not has_dbus:
            logger.warning("dbus-fast must be installed to listen to resume signals")
        self.resume = True

    def start(self) -> None:
        """
        Will create connection to dbus only if we want to listen out
        for a sleep or wake signal.
        """
        if not has_dbus:
            return

        if not (self.sleep or self.resume):
            return

        create_task(self._start())

    async def _start(self) -> None:
        """
        Creates the bus connection and connects to the org.freedesktop.login1.Manager
        interface. Starts an inhibitor if we are listening for sleep events.
        Attaches handler to the "PrepareForSleep" signal.
        """
        # Connect to bus and Manager interface
        try:
            self.bus = await MessageBus(bus_type=BusType.SYSTEM, negotiate_unix_fd=True).connect()
        except FileNotFoundError:
            self.bus = None
            logger.warning(
                "Could not find logind service. Suspend and resume hooks will be unavailable."
            )
            return

        try:
            introspection = await self.bus.introspect(LOGIND_SERVICE, LOGIND_PATH)
        except DBusError:
            logger.warning(
                "Could not find logind service. Suspend and resume hooks will be unavailable."
            )
            self.bus.disconnect()
            self.bus = None
            return

        obj = self.bus.get_proxy_object(LOGIND_SERVICE, LOGIND_PATH, introspection)
        self.login = obj.get_interface(LOGIND_INTERFACE)

        # If we want to know when the system is sleeping when we request an inhibitor
        if self.sleep:
            self.take()

        # Finally, attach a handler for the "PrepareForSleep" signal
        self.login.on_prepare_for_sleep(self.prepare_for_sleep)

    def take(self) -> None:
        """Create an inhibitor."""
        # Shouldn't happen but, if we already have an inhibitor in place,
        # close it before requesting a new one
        if self.fd >= 0:
            self.release()

        # Check that inhibitor was released
        if self.fd < 0:
            create_task(self._take())

    async def _take(self) -> None:
        """Sends the request to dbus to create an inhibitor."""
        # The "Inhibit" method returns a file descriptor
        self.fd = await self.login.call_inhibit(
            "sleep",  # what: The lock type. We only want to inhibit sleep
            "qtile",  # who: Name of program requesting inhibitor
            "Run hooked functions before suspend",  # why: Short description of purpose
            "delay",  # mode: "delay" or "block"
        )

        # We need to set CLOEXEC flag for the file descriptor
        # See: https://github.com/qtile/qtile/pull/4388#issuecomment-1675410090
        # for explanation
        flags = fcntl.fcntl(self.fd, fcntl.F_GETFD)
        fcntl.fcntl(self.fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

    def release(self) -> None:
        """Closes the file descriptor to release the inhibitor."""
        if self.fd >= 0:
            os.close(self.fd)

        try:
            os.fstat(self.fd)
        except OSError:
            # File descriptor was successfully closed
            self.fd = -1
        else:
            # We could read the file descriptor so it's still open
            logger.warning("Unable to release inhibitor.")

    def prepare_for_sleep(self, start: bool) -> None:
        """
        Handler for "PrepareForSleep" signal.

        Value of "sleep" is:
        - True when the machine is about to sleep
        - False when the event is over i.e. the machine has woken up
        """
        if start:
            hook.fire("suspend")

            # Note: lock is released after the "suspend" hook has been fired
            # Hooked functions should therefore be synchronous to ensure they
            # complete before the inhbitor is released.
            self.release()
        else:
            # If we're listening for suspend events, we need to request a new
            # inhibitor
            if self.sleep:
                self.take()
            hook.fire("resume")

    def stop(self) -> None:
        """
        Deactivates the inhibitor, removing lock and signal handler
        before closing bus connection.
        """
        if not has_dbus or self.bus is None:
            return

        if self.fd >= 0:
            self.release()

        if self.sleep or self.resume:
            self.login.off_prepare_for_sleep(self.prepare_for_sleep)

        self.bus.disconnect()
        self.bus = None


inhibitor = Inhibitor()



from asyncio import current_task
from collections.abc import Callable
from contextlib import suppress
from functools import partial
from pathlib import Path

# dbus_fast is incompatible with deferred type evaluation
import cairocffi
from dbus_fast import InterfaceNotFoundError, InvalidBusNameError, InvalidObjectPathError
from dbus_fast.aio import MessageBus
from dbus_fast.constants import PropertyAccess
from dbus_fast.errors import DBusError
from dbus_fast.service import ServiceInterface, dbus_property, method, signal

try:
    from xdg.IconTheme import getIconPath

    has_xdg = True
except ImportError:
    has_xdg = False

from libqtile.images import Img
from libqtile.log_utils import logger
from libqtile.utils import add_signal_receiver, create_task

ICON_FORMATS = [".png", ".svg"]

# StatusNotifier seems to have two potential interface names.
# While KDE appears to be the default, we should also listen
# for items on freedesktop.
BUS_NAMES = ["org.kde.StatusNotifierWatcher", "org.freedesktop.StatusNotifierWatcher"]

ITEM_INTERFACES = ["org.kde.StatusNotifierItem", "org.freedesktop.StatusNotifierItem"]

STATUSNOTIFIER_PATH = "/StatusNotifierItem"
PROTOCOL_VERSION = 0

STATUS_NOTIFIER_ITEM_SPEC = """
<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
"http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
<node>
  <interface name='org.kde.StatusNotifierItem'>
    <annotation name="org.gtk.GDBus.C.Name" value="Item" />
    <method name='ContextMenu'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='Activate'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='SecondaryActivate'>
      <arg type='i' direction='in' name='x'/>
      <arg type='i' direction='in' name='y'/>
    </method>
    <method name='Scroll'>
      <arg type='i' direction='in' name='delta'/>
      <arg type='s' direction='in' name='orientation'/>
    </method>
    <signal name='NewTitle'/>
    <signal name='NewIcon'/>
    <signal name='NewAttentionIcon'/>
    <signal name='NewOverlayIcon'/>
    <signal name='NewToolTip'/>
    <signal name='NewStatus'>
      <arg type='s' name='status'/>
    </signal>
    <property name='Category' type='s' access='read'/>
    <property name='Id' type='s' access='read'/>
    <property name='Title' type='s' access='read'/>
    <property name='Status' type='s' access='read'/>
    <!-- See discussion on pull #536
    <property name='WindowId' type='u' access='read'/>
    -->
    <property name='IconThemePath' type='s' access='read'/>
    <property name='IconName' type='s' access='read'/>
    <property name='IconPixmap' type='a(iiay)' access='read'/>
    <property name='OverlayIconName' type='s' access='read'/>
    <property name='OverlayIconPixmap' type='a(iiay)' access='read'/>
    <property name='AttentionIconName' type='s' access='read'/>
    <property name='AttentionIconPixmap' type='a(iiay)' access='read'/>
    <property name='AttentionMovieName' type='s' access='read'/>
    <property name='ToolTip' type='(sa(iiay)ss)' access='read'/>
    <property name='Menu' type='o' access='read'/>
    <property name='ItemIsMenu' type='b' access='read'/>
  </interface>
</node>
"""


class StatusNotifierItem:  # noqa: E303
    """
    Class object which represents an StatusNotiferItem object.

    The item is responsible for interacting with the
    application.
    """

    icon_map = {
        "Icon": ("_icon", "get_icon_pixmap"),
        "Attention": ("_attention_icon", "get_attention_icon_pixmap"),
        "Overlay": ("_overlay_icon", "get_overlay_icon_pixmap"),
    }

    def __init__(self, bus, service, path=None, icon_theme=None):
        self.bus = bus
        self.service = service
        self.surfaces = {}
        self._pixmaps = {}
        self._icon = None
        self._overlay_icon = None
        self._attention_icon = None
        self.on_icon_changed = None
        self.icon_theme = icon_theme
        self.icon = None
        self.path = path if path else STATUSNOTIFIER_PATH
        self.last_icon_name = None

    def __eq__(self, other):
        # Convenience method to find Item in list by service path
        if isinstance(other, StatusNotifierItem):
            return other.service == self.service
        elif isinstance(other, str):
            return other == self.service
        else:
            return False

    async def start(self):
        # Create a proxy object connecting for the item.
        # Some apps provide the incorrect path to the StatusNotifier object
        # We can try falling back to the default if that fails.
        # See: https://github.com/qtile/qtile/issues/3418
        # Note: this loop will run a maximum of two times and returns False
        # if the no object is available.
        found_path = False

        while not found_path:
            try:
                introspection = await self.bus.introspect(self.service, self.path)
                found_path = True
            except InvalidBusNameError:
                # This is probably an Ayatana indicator which doesn't provide the service name.
                # We'll pick it up via the message handler so we can ignore this.
                return False
            except InvalidObjectPathError:
                logger.info("Cannot find %s path on %s.", self.path, self.service)
                if self.path == STATUSNOTIFIER_PATH:
                    return False

                # Try the default ('/StatusNotifierItem')
                self.path = STATUSNOTIFIER_PATH

        try:
            obj = self.bus.get_proxy_object(self.service, self.path, introspection)
        except InvalidBusNameError:
            return False

        # Try to connect to the bus object and verify there's a valid
        # interface available
        # TODO: This may not ever fail given we've specified the underying
        # schema so dbus-fast has not attempted any introspection.
        interface_found = False
        for interface in ITEM_INTERFACES:
            try:
                self.item = obj.get_interface(interface)
                interface_found = True
                break
            except InterfaceNotFoundError:
                continue

        if not interface_found:
            logger.info(
                "Unable to find StatusNotifierItem interface on %s. Falling back to default spec.",
                self.service,
            )
            try:
                obj = self.bus.get_proxy_object(
                    self.service, STATUSNOTIFIER_PATH, STATUS_NOTIFIER_ITEM_SPEC
                )
                self.item = obj.get_interface("org.kde.StatusNotifierItem")
            except InterfaceNotFoundError:
                logger.warning(
                    "Failed to find StatusNotifierItem interface on %s and fallback to default spec also failed.",
                    self.service,
                )
                return False

        # Trying to get the local icon (first without fallback because there might be application-provided icons)
        await self._get_local_icon(fallback=False)

        # If there's no XDG icon, try to use icon provided by application
        if self.icon:
            self.item.on_new_icon(self._update_local_icon)
        else:
            # Get initial application icons:
            for icon in ["Icon", "Attention", "Overlay"]:
                await self._get_icon(icon)

            if self.has_icons:
                # Attach listeners for when the icon is updated
                self.item.on_new_icon(self._new_icon)
                self.item.on_new_attention_icon(self._new_attention_icon)
                self.item.on_new_overlay_icon(self._new_overlay_icon)

        if not self.has_icons:
            logger.warning(
                f"Cannot find icon: {self.last_icon_name!r} in current theme and no icon provided by StatusNotifierItem."
            )
            # No "local" icon and no application-provided icons are available.
            # The "local" icon may be updated at a later time, so "_update_local_icon"
            # gets registered for "on_new_icon" with the option to fall back to
            # a default icon.
            self.item.on_new_icon(self._update_local_icon)
            await self._get_local_icon()

        return True

    async def _get_local_icon(self, fallback=True):
        # Default to XDG icon
        # Some implementations don't provide an IconName property so we
        # need to catch an error if we can't read it.
        # We can't use hasattr to check this as the method will be created
        # where we've used the default XML spec to provide the object introspection
        icon_name = ""
        try:
            icon_name = await self.item.get_icon_name()
        except DBusError:
            return

        # We only need to do these searches if there's an icon name provided by
        # the app. We also don't want to do the recursive lookup with an empty
        # icon name as, otherwise, the glob will match things that are not images.
        if icon_name:
            if icon_name == self.last_icon_name:
                current_task().remove_done_callback(self._redraw)
                return

            self.last_icon_name = icon_name
            self.icon = None

            try:
                icon_path = await self.item.get_icon_theme_path()
            except (AttributeError, DBusError):
                icon_path = None

            icon = None
            if icon_path:
                icon = self._get_custom_icon(icon_name, Path(icon_path))

            if icon:
                self.icon = icon
            else:
                self.icon = self._get_xdg_icon(icon_name)
        else:
            self.icon = None

        if fallback:
            for icon in ["Icon", "Attention", "Overlay"]:
                await self._get_icon(icon)

        if not self.has_icons and fallback:
            # Use fallback icon libqtile/resources/status_notifier/fallback_icon.png
            logger.warning("Could not find icon for '%s'. Using fallback icon.", icon_name)
            path = Path(__file__).parent / "fallback_icon.png"
            self.icon = Img.from_path(path.resolve().as_posix())

    def _create_task_and_draw(self, coro):
        task = create_task(coro)
        task.add_done_callback(self._redraw)

    def _update_local_icon(self):
        self._create_task_and_draw(self._get_local_icon())

    def _new_icon(self):
        self._create_task_and_draw(self._get_icon("Icon"))

    def _new_attention_icon(self):
        self._create_task_and_draw(self._get_icon("Attention"))

    def _new_overlay_icon(self):
        self._create_task_and_draw(self._get_icon("Overlay"))

    def _get_custom_icon(self, icon_name, icon_path):
        icon = None
        for ext in ICON_FORMATS:
            path = icon_path / f"{icon_name}{ext}"
            if path.is_file():
                icon = path
                break

        else:
            # No icon found at the image path, let's search recursively
            glob = icon_path.rglob(f"{icon_name}.*")
            found = [
                icon for icon in glob if icon.is_file() and icon.suffix.lower() in ICON_FORMATS
            ]

            # Found a matching icon in subfolder
            if found:
                # We'd prefer an svg file
                svg = [icon for icon in found if icon.suffix.lower() == ".svg"]
                if svg:
                    icon = svg[0]
                else:
                    # If not, we'll take what there is
                    # NOTE: not clear how we can handle multiple matches with different icon sizes 16x16, 32x32 etc
                    icon = found[0]

        if icon is not None:
            return Img.from_path(icon.resolve().as_posix())

        return None

    def _get_xdg_icon(self, icon_name):
        if not has_xdg:
            return

        path = getIconPath(icon_name, theme=self.icon_theme, extensions=["png", "svg"])

        if not path:
            return None

        return Img.from_path(path)

    async def _get_icon(self, icon_name):
        """
        Requests the pixmap for the given `icon_name` and
        adds to an internal dictionary for later retrieval.
        """
        attr, method = self.icon_map[icon_name]
        pixmap = getattr(self.item, method, None)
        if pixmap is None:
            return
        icon_pixmap = await pixmap()

        # Items can present multiple pixmaps for different
        # size of icons. We want to keep these so we can pick
        # the best size when redering the icon later.
        # Also, the bytes sent for the pixmap are big-endian
        # but Cairo expects little-endian so we need to
        # reorder them.
        self._pixmaps[icon_name] = {
            size: self._reorder_bytes(icon_bytes) for size, _, icon_bytes in icon_pixmap
        }

    def _reorder_bytes(self, icon_bytes):
        """
        Method loops over the array and reverses every
        4 bytes (representing one RGBA pixel).
        """
        arr = bytearray(icon_bytes)
        for i in range(0, len(arr), 4):
            arr[i : i + 4] = arr[i : i + 4][::-1]

        return arr

    def _redraw(self, result):
        """Method to invalidate icon cache and redraw icons."""
        self._invalidate_icons()
        if self.on_icon_changed is not None:
            self.on_icon_changed(self)

    def _invalidate_icons(self):
        self.surfaces = {}

    def _get_sizes(self):
        """Returns list of available icon sizes."""
        if not self._pixmaps.get("Icon", False):
            return []

        return sorted([size for size in self._pixmaps["Icon"]])

    def _get_surfaces(self, size):
        """
        Creates a Cairo ImageSurface for each available icon
        for the given size.
        """
        raw_surfaces = {}
        for icon in self._pixmaps:
            if size in self._pixmaps[icon]:
                srf = cairocffi.ImageSurface.create_for_data(
                    self._pixmaps[icon][size], cairocffi.FORMAT_ARGB32, size, size
                )
                raw_surfaces[icon] = srf
        return raw_surfaces

    def get_icon(self, size):
        """
        Returns a cairo ImageSurface for the selected `size`.

        Will pick the appropriate icon and add any overlay as required.
        """
        # Use existing icon if generated previously
        if size in self.surfaces:
            return self.surfaces[size]

        # Create a blank ImageSurface to hold the icon
        icon = cairocffi.ImageSurface(cairocffi.FORMAT_ARGB32, size, size)

        if self.icon:
            base_icon = self.icon.surface
            icon_size = base_icon.get_width()
            overlay = None

        else:
            # Find best matching icon size:
            # We get all available sizes and filter this list so it only shows
            # the icon sizes bigger than the requested size (we prefer to
            # shrink icons rather than scale them up)
            all_sizes = self._get_sizes()
            sizes = [s for s in all_sizes if s >= size]

            # TODO: This is messy. Shouldn't return blank icon
            # If there are no sizes at all (i.e. no icon) then we return empty
            # icon
            if not all_sizes:
                return icon

            # Choose the first available size. If there are none (i.e. we
            # request icon size bigger than the largest provided by the app),
            # we just take the largest icon
            icon_size = sizes[0] if sizes else all_sizes[-1]

            srfs = self._get_surfaces(icon_size)

            # TODO: This shouldn't happen...
            if not srfs:
                return icon

            # TODO: Check spec for when to use "attention"
            base_icon = srfs.get("Attention", srfs["Icon"])
            overlay = srfs.get("Overlay", None)

        with cairocffi.Context(icon) as ctx:
            scale = size / icon_size
            ctx.scale(scale, scale)
            ctx.set_source_surface(base_icon)
            ctx.paint()
            if overlay:
                ctx.set_source_surface(overlay)
                ctx.paint()

        # Store the surface for next time
        self.surfaces[size] = icon

        return icon

    def activate(self):
        if hasattr(self.item, "call_activate"):
            create_task(self._activate())

    async def _activate(self):
        # Call Activate method and pass window position hints
        await self.item.call_activate(0, 0)

    @property
    def has_icons(self):
        return any(bool(icon) for icon in self._pixmaps.values()) or self.icon is not None


class StatusNotifierWatcher(ServiceInterface):  # noqa: E303
    """
    DBus service that creates a StatusNotifierWatcher interface
    on the bus and listens for applications wanting to register
    items.
    """

    def __init__(self, service: str):
        super().__init__(service)
        self._items: list[str] = []
        self._hosts: list[str] = []
        self.service = service
        self.on_item_added: Callable | None = None
        self.on_host_added: Callable | None = None
        self.on_item_removed: Callable | None = None
        self.on_host_removed: Callable | None = None

    async def start(self):
        # Set up and register the service on ths bus
        self.bus = await MessageBus().connect()
        self.bus.add_message_handler(self._message_handler)
        self.bus.export("/StatusNotifierWatcher", self)
        await self.bus.request_name(self.service)

        # We need to listen for interfaces being removed from
        # the bus so we can remove icons when the application
        # is closed.
        await self._setup_listeners()

    def _message_handler(self, message):
        """
        Low level method to check incoming messages.

        Ayatana indicators seem to register themselves by passing their object
        path rather than the service providing that object. We therefore need
        to identify the sender of the message in order to register the service.

        Returning False so senders receieve a reply (returning True prevents
        reply being sent)
        """
        if message.member != "RegisterStatusNotifierItem":
            return False

        # If the argument is not an object path (starting with "/") then we assume
        # it is the bus name and we don't need to do anything else.
        if not message.body[0].startswith("/"):
            return False

        if message.sender not in self._items:
            self._items.append(message.sender)
            if self.on_item_added is not None:
                self.on_item_added(message.sender, message.body[0])
            self.StatusNotifierItemRegistered(message.sender)
        return False

    async def _setup_listeners(self):
        """
        Register a MatchRule to receive signals when interfaces are added
        and removed from the bus.
        """
        await add_signal_receiver(
            self._name_owner_changed,
            session_bus=True,
            signal_name="NameOwnerChanged",
            dbus_interface="org.freedesktop.DBus",
            use_bus=self.bus,
            preserve=True,
        )

    def _name_owner_changed(self, message):
        # We need to track when an interface has been removed from the bus
        # We use the NameOwnerChanged signal and check if the new owner is
        # empty.
        name, _, new_owner = message.body

        # Check if one of our registered items or hosts has been removed.
        # If so, remove from our list and emit relevant signal
        if new_owner == "" and name in self._items:
            self._items.remove(name)
            self.StatusNotifierItemUnregistered(name)

        if new_owner == "" and name in self._hosts:
            self._hosts.remove(name)
            self.StatusNotifierHostUnregistered(name)

    @method()
    def RegisterStatusNotifierItem(self, service: "s"):  # type: ignore  # noqa: F821, N802
        if service not in self._items:
            self._items.append(service)
            if self.on_item_added is not None:
                self.on_item_added(service)
            self.StatusNotifierItemRegistered(service)

    @method()
    def RegisterStatusNotifierHost(self, service: "s"):  # type: ignore  # noqa: F821, N802
        if service not in self._hosts:
            self._hosts.append(service)
            self.StatusNotifierHostRegistered(service)

    @dbus_property(access=PropertyAccess.READ)
    def RegisteredStatusNotifierItems(self) -> "as":  # type: ignore  # noqa: F722, F821, N802
        return self._items

    @dbus_property(access=PropertyAccess.READ)
    def IsStatusNotifierHostRegistered(self) -> "b":  # type: ignore  # noqa: F821, N802
        # Note: applications may not register items unless this
        # returns True
        return len(self._hosts) > 0

    @dbus_property(access=PropertyAccess.READ)
    def ProtocolVersion(self) -> "i":  # type: ignore  # noqa: F821, N802
        return PROTOCOL_VERSION

    @signal()
    def StatusNotifierItemRegistered(self, service) -> "s":  # type: ignore  # noqa: F821, N802
        return service

    @signal()
    def StatusNotifierItemUnregistered(self, service) -> "s":  # type: ignore  # noqa: F821, N802
        if self.on_item_removed is not None:
            self.on_item_removed(service)
        return service

    @signal()
    def StatusNotifierHostRegistered(self, service) -> "s":  # type: ignore  # noqa: F821, N802
        if self.on_host_added is not None:
            self.on_host_added(service)
        return service

    @signal()
    def StatusNotifierHostUnregistered(self, service) -> "s":  # type: ignore  # noqa: F821, N802
        if self.on_host_removed is not None:
            self.on_host_removed(service)
        return service


class StatusNotifierHost:  # noqa: E303
    """
    Host object to act as a bridge between the widget and the DBus objects.

    The Host collates items returned from multiple watcher interfaces and
    collates them into a single list for the widget to access.
    """

    def __init__(self):
        self.watchers: list[StatusNotifierWatcher] = []
        self.items: list[StatusNotifierItem] = []
        self.name = "qtile"
        self.icon_theme: str = None
        self.started = False
        self._on_item_added: list[Callable] = []
        self._on_item_removed: list[Callable] = []
        self._on_icon_changed: list[Callable] = []

    async def start(
        self,
        on_item_added: Callable | None = None,
        on_item_removed: Callable | None = None,
        on_icon_changed: Callable | None = None,
    ):
        """
        Starts the host if not already started.

        Widgets should register their callbacks via this method.
        """
        if on_item_added:
            self._on_item_added.append(on_item_added)

        if on_item_removed:
            self._on_item_removed.append(on_item_removed)

        if on_icon_changed:
            self._on_icon_changed.append(on_icon_changed)

        if self.started:
            if on_item_added:
                for item in self.items:
                    on_item_added(item)

        else:
            self.bus = await MessageBus().connect()
            await self.bus.request_name("org.freedesktop.StatusNotifierHost-qtile")
            for iface in BUS_NAMES:
                w = StatusNotifierWatcher(iface)
                w.on_item_added = self.add_item
                w.on_item_removed = self.remove_item
                await w.start()

                # Not quite following spec here as we're not registering
                # the host on the bus.
                w.RegisterStatusNotifierHost(self.name)
                self.watchers.append(w)
            self.started = True

    def item_added(self, item, service, future):
        success = future.result()
        # If StatusNotifierItem object was created successfully then we
        # add to our list and redraw the bar
        if success:
            self.items.append(item)
            for callback in self._on_item_added:
                callback(item)

        # It's an invalid item so let's remove it from the watchers
        else:
            for w in self.watchers:
                try:
                    w._items.remove(service)
                except ValueError:
                    pass

    def add_item(self, service, path=None):
        """
        Creates a StatusNotifierItem for the given service and tries to
        start it.
        """
        item = StatusNotifierItem(self.bus, service, path=path, icon_theme=self.icon_theme)
        item.on_icon_changed = self.item_icon_changed
        if item not in self.items:
            task = create_task(item.start())
            task.add_done_callback(partial(self.item_added, item, service))

    def remove_item(self, interface):
        # Check if the interface is in out list of items and, if so,
        # remove it and redraw the bar
        if interface in self.items:
            self.items.remove(interface)
            for callback in self._on_item_removed:
                callback(interface)

    def item_icon_changed(self, item):
        for callback in self._on_icon_changed:
            callback(item)

    def unregister_callbacks(
        self, on_item_added=None, on_item_removed=None, on_icon_changed=None
    ):
        if on_item_added is not None:
            with suppress(ValueError):
                self._on_item_added.remove(on_item_added)

        if on_item_removed is not None:
            with suppress(ValueError):
                self._on_item_removed.remove(on_item_removed)

        if on_icon_changed is not None:
            with suppress(ValueError):
                self._on_icon_changed.remove(on_icon_changed)


host = StatusNotifierHost()  # noqa: E303



from libqtile import bar
from libqtile.widget import base


class _CrashMe(base._TextBox):
    """A developer widget to force a crash in qtile

    Pressing left mouse button causes a zero divison error.  Pressing the right
    mouse button causes a cairo draw error.

    Parameters
    ==========

    width :
        A fixed width, or bar.CALCULATED to calculate the width automatically
        (which is recommended).
    """

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "Crash me !", width, **config)

    def _configure(self, qtile, bar):
        base._Widget._configure(self, qtile, bar)
        self.layout = self.drawer.textlayout(
            self.text, self.foreground, self.font, self.fontsize, self.fontshadow, markup=True
        )

    def button_press(self, x, y, button):
        if button == 1:
            1 / 0



import xcffib.screensaver

from libqtile.backend.base.idle_notify import IdleNotifier as BaseIdleNotifier
from libqtile.log_utils import logger


class IdleNotifier(BaseIdleNotifier):
    def __init__(self, core):
        super().__init__(core)
        self.has_screen_saver = hasattr(self.core.conn, "mit_screen_saver")

    @property
    def timeout_increments(self):
        timers = self.timeouts
        return [timers[0]] + [timers[i] - timers[i - 1] for i in range(1, len(timers))]

    @property
    def has_timeout(self):
        return self.timeouts and self.index < len(self.timeouts)

    def run(self):
        if not self.has_screen_saver:
            logger.warning(
                "Idle timers need the screensaver extension for x11. Timers will not be run."
            )
            return

        if not self.timeouts:
            return

        self.core.conn.mit_screen_saver.select_events(self.core._root.wid)
        self.reset()

    def reset(self):
        self.index = 0
        if self.has_timeout:
            self.set_timer()

    def set_timer(self):
        interval = self.timeout_increments[self.index]
        self.core.conn.mit_screen_saver.set_interval(interval)

    def handle_timeout(self):
        timeout = self.timeouts[self.index]
        self.fire_action(timeout)

        self.index += 1
        if self.has_timeout:
            self.set_timer()

    def handle_resume(self):
        if self.index > 0:
            self.fire_resume()

        self.reset()

    def check_event(self, event) -> bool:
        if not self.has_screen_saver:
            return False

        if type(event) is not xcffib.screensaver.NotifyEvent:
            return False

        if self.has_timeout and event.state in (
            xcffib.screensaver.State.On,
            xcffib.screensaver.State.Cycle,
        ):
            self.handle_timeout()
        elif event.state == xcffib.screensaver.State.Off:
            self.handle_resume()

        return True

    def clear_timers(self) -> None:
        super().clear_timers()
        self.reset()



from libqtile.log_utils import logger
from libqtile.widget.generic_poll_text import GenPollCommand


class Canto(GenPollCommand):
    """Display RSS feeds updates using the canto remote

    Widget requirements: canto_

    .. _canto: https://codezen.org/canto-ng/
    """

    defaults = [
        ("tags", [], "List of tags to display, empty for all"),
        ("one_format", "{name}: {number}", "One tag display format"),
        ("all_format", "{number}", "All tags display format"),
    ]

    def __init__(self, **config):
        config["cmd"] = ["canto-remote", "status", "--tags"]
        GenPollCommand.__init__(self, **config)
        self.add_defaults(Canto.defaults)

    def get_info(self, output):
        output = output.splitlines()
        if not self.tags:
            total_items = 0
            for line in output:
                if "maintag:" in line:
                    current_tag_items = line[line.index(":", line.index(":") + 1) + 2 :]
                    total_items += int(current_tag_items)
            return total_items
        else:
            all_tags_output = {}
            for line in output:
                second_colon_index = line.index(":", line.index(":") + 1)
                current_tag_name = line[:second_colon_index].strip()
                if current_tag_name in self.tags:
                    current_tag_items = line[second_colon_index + 2 :].strip()
                    all_tags_output.update({current_tag_name: current_tag_items})
            return all_tags_output

    def parse(self, output):
        output = self.get_info(output)
        if not self.tags:
            display = self.all_format.format(number=output)
            return display
        else:
            if not output:
                logger.debug("Canto remote found no tags")
                return ""
            display = []
            for key in output:
                current_tag_output = self.one_format.format(name=key, number=output[key])
                display.append(current_tag_output)
            return "".join(display)



import csv
import re

from libqtile.widget import base

sensors_mapping = {
    "fan_speed": "fan.speed",
    "perf": "pstate",
    "temp": "temperature.gpu",
}


def _all_sensors_names_correct(sensors):
    return all(map(lambda x: x in sensors_mapping, sensors))


class NvidiaSensors(base.BackgroundPoll):
    """Displays temperature, fan speed and performance level Nvidia GPU."""

    defaults = [
        (
            "format",
            "{temp}°C",
            "Display string format. Three options available:  "
            "``{temp}`` - temperature, ``{fan_speed}`` and ``{perf}`` - "
            "performance level",
        ),
        ("foreground_alert", "ff0000", "Foreground colour alert"),
        (
            "gpu_bus_id",
            "",
            "GPU's Bus ID, ex: ``01:00.0``. If leave empty will display all available GPU's",
        ),
        ("update_interval", 2, "Update interval in seconds."),
        (
            "threshold",
            70,
            "If the current temperature value is above, then change to foreground_alert colour",
        ),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(NvidiaSensors.defaults)
        self.foreground_normal = self.foreground

    def _get_sensors_data(self, command):
        return csv.reader(
            self.call_process(command, shell=True).strip().replace(" ", "").split("\n")
        )

    def _parse_format_string(self):
        return {sensor for sensor in re.findall("{(.+?)}", self.format)}

    def poll(self):
        sensors = self._parse_format_string()
        if not _all_sensors_names_correct(sensors):
            return "Wrong sensor name"
        bus_id = f"-i {self.gpu_bus_id}" if self.gpu_bus_id else ""
        command = "nvidia-smi {} --query-gpu={} --format=csv,noheader".format(
            bus_id, ",".join(sensors_mapping[sensor] for sensor in sensors)
        )
        try:
            sensors_data = [dict(zip(sensors, gpu)) for gpu in self._get_sensors_data(command)]
            for gpu in sensors_data:
                if gpu.get("temp"):
                    if int(gpu["temp"]) > self.threshold:
                        self.layout.colour = self.foreground_alert
                    else:
                        self.layout.colour = self.foreground_normal
            return " - ".join([self.format.format(**gpu) for gpu in sensors_data])
        except Exception:
            return None



import imaplib
import re

from libqtile.log_utils import logger
from libqtile.widget import base


class GmailChecker(base.BackgroundPoll):
    """
    A simple gmail checker. If 'status_only_unseen' is True - set 'fmt' for one
    argument, ex. 'unseen: {0}'
    """

    defaults = [
        ("update_interval", 30, "Update time in seconds."),
        ("username", None, "username"),
        ("password", None, "password"),
        ("email_path", "INBOX", "email_path"),
        ("display_fmt", "inbox[{0}],unseen[{1}]", "Display format"),
        ("status_only_unseen", False, "Only show unseen messages"),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(GmailChecker.defaults)

    def poll(self):
        self.gmail = imaplib.IMAP4_SSL("imap.gmail.com")
        self.gmail.login(self.username, self.password)
        answer, raw_data = self.gmail.status(self.email_path, "(MESSAGES UNSEEN)")
        if answer == "OK":
            dec = raw_data[0].decode()
            messages = int(re.search(r"MESSAGES\s+(\d+)", dec).group(1))
            unseen = int(re.search(r"UNSEEN\s+(\d+)", dec).group(1))
            if self.status_only_unseen:
                return self.display_fmt.format(unseen)
            else:
                return self.display_fmt.format(messages, unseen)
        else:
            logger.exception(
                "GmailChecker UNKNOWN error, answer: %s, raw_data: %s", answer, raw_data
            )
            return "UNKNOWN ERROR"



"""
A widget for Music Player Daemon (MPD) based on python-mpd2.

This widget exists since python-mpd library is no longer supported.
"""

from collections import defaultdict
from html import escape

from mpd import CommandError, ConnectionError, MPDClient

from libqtile import utils
from libqtile.log_utils import logger
from libqtile.widget import base

# Mouse Interaction
# TODO: Volume inc/dec support
keys = {
    # Left mouse button
    1: "toggle",
    # Right mouse button
    3: "stop",
    # Scroll up
    4: "previous",
    # Scroll down
    5: "next",
}

# To display mpd state
play_states = {"play": "\u25b6", "pause": "\u23f8", "stop": "\u25a0"}


def option(char):
    """
    old status mapping method.

    Deprecated.
    """

    def _convert(elements, key, space):
        if key in elements and elements[key] != "0":
            elements[key] = char
        else:
            elements[key] = space

    return _convert


# Changes to formatter will still use this dicitionary as a fallback
prepare_status = {
    "repeat": option("r"),
    "random": option("z"),
    "single": option("1"),
    "consume": option("c"),
    "updating_db": option("U"),
}

# dictionary for new formatting method.  This is now default.
status_dict = {"repeat": "r", "random": "z", "single": "1", "consume": "c", "updating_db": "U"}

default_idle_message = "MPD IDLE"

default_idle_format = (
    "{play_status} {idle_message}" + "[{repeat}{random}{single}{consume}{updating_db}]"
)

default_format = (
    "{play_status} {artist}/{title} " + "[{repeat}{random}{single}{consume}{updating_db}]"
)

default_undefined_status_value = "Undefined"


def default_cmd():
    return None


format_fns = {
    "all": escape,
}


class Mpd2(base.BackgroundPoll):
    r"""Mpd2 Object.

    Parameters
    ==========
    status_format:
        format string to display status

        For a full list of values, see:
            MPDClient.status() and MPDClient.currentsong()

        https://musicpd.org/doc/protocol/command_reference.html#command_status
        https://musicpd.org/doc/protocol/tags.html

        Default::

            '{play_status} {artist}/{title} \
                [{repeat}{random}{single}{consume}{updating_db}]'

            ``play_status`` is a string from ``play_states`` dict

            Note that the ``time`` property of the song renamed to ``fulltime``
            to prevent conflicts with status information during formating.

    idle_format:
        format string to display status when no song is in queue.

        Default::

            '{play_status} {idle_message} \
                [{repeat}{random}{single}{consume}{updating_db}]'

            Note that the ``artist`` key fallbacks to similar keys in specific order.
            (``artist`` -> ``albumartist`` -> ``performer`` ->
             -> ``composer`` -> ``conductor`` -> ``ensemble``)

    idle_message:
        text to display instead of song information when MPD is idle.
        (i.e. no song in queue)

        Default:: "MPD IDLE"

    undefined_value:
        text to display when status key is undefined

        Default:: "Undefined"

    prepare_status:
        dict of functions to replace values in status with custom characters.

        ``f(status, key, space_element) => str``

        New functionality allows use of a dictionary of plain strings.

        Default::

            status_dict = {
                'repeat': 'r',
                'random': 'z',
                'single': '1',
                'consume': 'c',
                'updating_db': 'U'
            }

    format_fns:
        A dict of functions to format the various elements.

        'Tag': f(str) => str

        Default:: { 'all': lambda s: cgi.escape(s) }

        N.B. if 'all' is present, it is processed on every element of song_info
            before any other formatting is done.

    mouse_buttons:
        A dict of mouse button numbers to actions

    Widget requirements: python-mpd2_.

    .. _python-mpd2: https://pypi.org/project/python-mpd2/
    """

    defaults = [
        ("update_interval", 1, "Interval of update widget"),
        ("host", "localhost", "Host of mpd server"),
        ("port", 6600, "Port of mpd server"),
        ("password", None, "Password for auth on mpd server"),
        ("mouse_buttons", keys, "b_num -> action."),
        ("play_states", play_states, "Play state mapping"),
        ("format_fns", format_fns, "Dictionary of format methods"),
        ("command", default_cmd, "command to be executed by mapped mouse button."),
        ("prepare_status", status_dict, "characters to show the status of MPD"),
        ("status_format", default_format, "format for displayed song info."),
        ("idle_format", default_idle_format, "format for status when mpd has no playlist."),
        ("idle_message", default_idle_message, "text to display when mpd is idle."),
        (
            "undefined_value",
            default_undefined_status_value,
            "text to display when status key is undefined.",
        ),
        ("timeout", 30, "MPDClient timeout"),
        ("idletimeout", 5, "MPDClient idle command timeout"),
        ("no_connection", "No connection", "Text when mpd is disconnected"),
        ("color_progress", None, "Text color to indicate track progress."),
        ("space", "-", "Space keeper"),
    ]

    def __init__(self, **config):
        """Constructor."""
        super().__init__("", **config)

        self.add_defaults(Mpd2.defaults)
        if self.color_progress:
            self.color_progress = utils.hex(self.color_progress)

    def _configure(self, qtile, bar):
        super()._configure(qtile, bar)
        self.client = MPDClient()
        self.client.timeout = self.timeout
        self.client.idletimeout = self.idletimeout

    @property
    def connected(self):
        """Attempt connection to mpd server."""
        try:
            self.client.ping()  # pylint: disable=E1101
        except (OSError, ConnectionError):
            try:
                self.client.connect(self.host, self.port)
                if self.password:
                    self.client.password(self.password)  # pylint: disable=E1101
            except (OSError, ConnectionError, CommandError):
                return False
        return True

    def poll(self):
        """
        Called by qtile manager.

        poll the mpd server and update widget.
        """
        if self.connected:
            return self.update_status()
        else:
            return self.no_connection

    def update_status(self):
        """get updated info from mpd server and call format."""
        self.client.command_list_ok_begin()
        self.client.status()  # pylint: disable=E1101
        self.client.currentsong()  # pylint: disable=E1101
        status, current_song = self.client.command_list_end()

        return self.formatter(status, current_song)

    def button_press(self, x, y, button):
        """handle click event on widget."""
        base.BackgroundPoll.button_press(self, x, y, button)
        m_name = self.mouse_buttons[button]

        if self.connected:
            if hasattr(self, m_name):
                self.__try_call(m_name)
            elif hasattr(self.client, m_name):
                self.__try_call(m_name, self.client)

    def __try_call(self, attr_name, obj=None):
        err1 = "Class {Class} has no attribute {attr}."
        err2 = 'attribute "{Class}.{attr}" is not callable.'
        context = obj or self
        try:
            getattr(context, attr_name)()
        except (AttributeError, TypeError) as e:
            if isinstance(e, AttributeError):
                err = err1.format(Class=type(context).__name__, attr=attr_name)
            else:
                err = err2.format(Class=type(context).__name__, attr=attr_name)
            logger.exception("%s %s", err, e.args[0])

    def toggle(self):
        """toggle play/pause."""
        status = self.client.status()  # pylint: disable=E1101
        play_status = status["state"]

        if play_status == "play":
            self.client.pause()  # pylint: disable=E1101
        else:
            self.client.play()  # pylint: disable=E1101

    def formatter(self, status, current_song):
        """format song info."""
        song_info = defaultdict(lambda: self.undefined_value)
        song_info["play_status"] = self.play_states[status["state"]]

        if status["state"] == "stop" and current_song == {}:
            song_info["idle_message"] = self.idle_message
            fmt = self.idle_format
        else:
            fmt = self.status_format

        for k in current_song:
            song_info[k] = current_song[k]
        song_info["fulltime"] = song_info["time"]
        del song_info["time"]

        song_info.update(status)
        if song_info["updating_db"] == self.undefined_value:
            song_info["updating_db"] = "0"
        if not callable(self.prepare_status["repeat"]):
            for k in self.prepare_status:
                if k in status and status[k] != "0":
                    # Much more direct.
                    song_info[k] = self.prepare_status[k]
                else:
                    song_info[k] = self.space
        else:
            self.prepare_formatting(song_info)

        # 'remaining' isn't actually in the information provided by mpd
        # so we construct it from 'fulltime' and 'elapsed'.
        # 'elapsed' is always less than or equal to 'fulltime', if it exists.
        # Remaining should default to '00:00' if either or both are missing.
        # These values are also used for coloring text by progress, if wanted.
        if "remaining" in self.status_format or self.color_progress:
            total = (
                float(song_info["fulltime"])
                if song_info["fulltime"] != self.undefined_value
                else 0.0
            )
            elapsed = (
                float(song_info["elapsed"])
                if song_info["elapsed"] != self.undefined_value
                else 0.0
            )
            song_info["remaining"] = f"{float(total - elapsed):.2f}"

        if "song" in self.status_format and song_info["song"] != self.undefined_value:
            song_info["currentsong"] = str(int(song_info["song"]) + 1)

        if "artist" in self.status_format and song_info["artist"] == self.undefined_value:
            artist_keys = ("albumartist", "performer", "composer", "conductor", "ensemble")
            for key in artist_keys:
                if song_info[key] != self.undefined_value:
                    song_info["artist"] = song_info[key]
                    break

        # mpd serializes tags containing commas as lists.
        for key in song_info:
            if isinstance(song_info[key], list):
                song_info[key] = ", ".join(song_info[key])

        # Now we apply the user formatting to selected elements in song_info.
        # if 'all' is defined, it is applied first.
        # the reason for this is that, if the format functions do pango markup.
        # we don't want to do anything that would mess it up, e.g. `escape`ing.
        if "all" in self.format_fns:
            for key in song_info:
                song_info[key] = self.format_fns["all"](song_info[key])
        for fmt_fn in self.format_fns:
            if fmt_fn in song_info and fmt_fn != "all":
                song_info[fmt_fn] = self.format_fns[fmt_fn](song_info[fmt_fn])

        # fmt = self.status_format
        if not isinstance(fmt, str):
            fmt = str(fmt)

        formatted = fmt.format_map(song_info)

        if self.color_progress and status["state"] != "stop":
            try:
                progress = int(len(formatted) * elapsed / total)
                formatted = f'<span color="{self.color_progress}">{formatted[:progress]}</span>{formatted[progress:]}'
            except (ZeroDivisionError, ValueError):
                pass

        return formatted

    def prepare_formatting(self, status):
        """old way of preparing status formatting."""
        for key in self.prepare_status:
            self.prepare_status[key](status, key, self.space)

    def finalize(self):
        """finalize."""
        super().finalize()

        try:
            self.client.close()  # pylint: disable=E1101
            self.client.disconnect()
        except ConnectionError:
            pass



from ipykernel.kernelbase import Kernel

from libqtile import command, ipc, sh


class QshKernel(Kernel):
    implementation = "qshell"
    implementation_version = "0.1"
    language = "no-op"
    language_version = "1.0"
    language_info = {"mimetype": "text/plain"}
    banner = "Qsh Kernel"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        socket_path = ipc.find_sockfile()
        ipc_client = ipc.Client(socket_path)
        cmd_object = command.interface.IPCCommandInterface(ipc_client)
        self.qsh = sh.QSh(cmd_object)

    def do_execute(
        self, code, silent, _store_history=True, _user_expressions=None, _allow_stdin=False
    ):
        # if no command sent, just return
        if not code.strip():
            return {
                "status": "ok",
                "execution_count": self.execution_count,
                "payload": [],
                "user_expressions": {},
            }

        if code[-1] == "?":
            return self.do_inspect(code, len(code) - 1)

        try:
            output = self.qsh.process_line(code)
        except KeyboardInterrupt:
            return {
                "status": "abort",
                "execution_count": self.execution_count,
            }

        if not silent and output:
            stream_content = {"name": "stdout", "text": output}
            self.send_response(self.iopub_socket, "stream", stream_content)

        return {
            "status": "ok",
            "execution_count": self.execution_count,
            "payload": [],
            "user_expressions": {},
        }

    def do_complete(self, code, cursor_pos):
        no_complete = {
            "status": "ok",
            "matches": [],
            "cursor_start": 0,
            "cursor_end": cursor_pos,
            "metadata": dict(),
        }

        if not code or code[-1] == " ":
            return no_complete

        tokens = code.split()
        if not tokens:
            return no_complete

        token = tokens[-1]
        start = cursor_pos - len(token)

        matches = self.qsh._complete(code, token)
        return {
            "status": "ok",
            "matches": sorted(matches),
            "cursor_start": start,
            "cursor_end": cursor_pos,
            "metadata": dict(),
        }


def main():
    from ipykernel.kernelapp import IPKernelApp

    IPKernelApp.launch_instance(kernel_class=QshKernel)


if __name__ == "__main__":
    main()



import time
from typing import Any
from urllib.parse import urlencode

from libqtile.widget.gen_poll_url import GenPollUrl

# See documentation: https://openweathermap.org/current
QUERY_URL = "https://api.openweathermap.org/data/2.5/weather?"
DEFAULT_APP_ID = "7834197c2338888258f8cb94ae14ef49"


class OpenWeatherResponseError(Exception):
    def __init__(self, resp_code, err_str=None):
        self.resp_code = resp_code
        self.err_str = err_str


def flatten_json(obj):
    out = {}

    def __inner(_json, name=""):
        if type(_json) is dict:
            for key, value in _json.items():
                __inner(value, name + key + "_")
        elif type(_json) is list:
            for i in range(len(_json)):
                __inner(_json[i], name + str(i) + "_")
        else:
            out[name[:-1]] = _json

    __inner(obj)
    return out


def degrees_to_direction(degrees):
    val = int(degrees / 22.5 + 0.5)
    arr = [
        "N",
        "NNE",
        "NE",
        "ENE",
        "E",
        "ESE",
        "SE",
        "SSE",
        "S",
        "SSW",
        "SW",
        "WSW",
        "W",
        "WNW",
        "NW",
        "NNW",
    ]
    return arr[(val % 16)]


class _OpenWeatherResponseParser:
    def __init__(self, response, dateformat, timeformat):
        self.dateformat = dateformat
        self.timeformat = timeformat
        self.data = self._parse(response)
        self._remap(self.data)
        if int(self.data["cod"]) != 200:
            raise OpenWeatherResponseError(int(self.data["cod"]))

    def _parse(self, response):
        return flatten_json(response)

    def _remap(self, data):
        data["location_lat"] = data.get("coord_lat", None)
        data["location_long"] = data.get("coord_lon", None)
        data["location_city"] = data.get("name", None)
        data["location_cityid"] = data.get("id", None)
        data["location_country"] = data.get("sys_country", None)
        data["sunrise"] = self._get_sunrise_time()
        data["sunset"] = self._get_sunset_time()
        data["isotime"] = self._get_dt()
        data["wind_direction"] = self._get_wind_direction()
        data["weather"] = data.get("weather_0_main", None)
        data["weather_details"] = data.get("weather_0_description", None)
        data["humidity"] = data.get("main_humidity", None)
        data["pressure"] = data.get("main_pressure", None)
        data["temp"] = data.get("main_temp", None)

    def _get_wind_direction(self):
        wd = self.data.get("wind_deg", None)
        if wd is None:
            return None
        return degrees_to_direction(wd)

    def _get_sunrise_time(self):
        dt = self.data.get("sys_sunrise", None)
        if dt is None:
            return None
        return time.strftime(self.timeformat, time.localtime(dt))

    def _get_sunset_time(self):
        dt = self.data.get("sys_sunset", None)
        if dt is None:
            return None
        return time.strftime(self.timeformat, time.localtime(dt))

    def _get_dt(self):
        dt = self.data.get("dt", None)
        if dt is None:
            return None
        return time.strftime(self.dateformat + self.timeformat, time.localtime(dt))


class OpenWeather(GenPollUrl):
    """A weather widget, data provided by the OpenWeather API.

    Some format options:
        - location_city
        - location_cityid
        - location_country
        - location_lat
        - location_long

        - weather
        - weather_details
        - units_temperature
        - units_wind_speed
        - isotime
        - humidity
        - pressure
        - sunrise
        - sunset
        - temp
        - visibility
        - wind_speed
        - wind_deg
        - wind_direction

        - main_feels_like
        - main_temp_min
        - main_temp_max
        - clouds_all
        - icon

    Icon support is available but you will need a suitable font installed. A default icon mapping
    is provided (``OpenWeather.symbols``) but changes can be made by setting ``weather_symbols``.
    Available icon codes can be viewed here: https://openweathermap.org/weather-conditions#Icon-list
    """

    symbols = {
        "Unknown": "✨",
        "01d": "☀️",
        "01n": "🌕",
        "02d": "🌤️",
        "02n": "☁️",
        "03d": "🌥️",
        "03n": "☁️",
        "04d": "☁️",
        "04n": "☁️",
        "09d": "🌧️",
        "09n": "🌧️",
        "10d": "⛈",
        "10n": "⛈",
        "11d": "🌩",
        "11n": "🌩",
        "13d": "❄️",
        "13n": "❄️",
        "50d": "🌫",
        "50n": "🌫",
    }

    defaults: list[tuple[str, Any, str]] = [
        # One of (cityid, location, zip, coordinates) must be set.
        (
            "app_key",
            DEFAULT_APP_ID,
            """Open Weather access key. A default is provided, but
            for prolonged use obtaining your own is suggested:
            https://home.openweathermap.org/users/sign_up""",
        ),
        (
            "cityid",
            None,
            """City ID. Can be looked up on e.g.:
            https://openweathermap.org/find
            Takes precedence over location and coordinates.
            Note that this is not equal to a WOEID.""",
        ),
        (
            "location",
            None,
            """Name of the city. Country name can be appended
            like cambridge,NZ. Takes precedence over zip-code.""",
        ),
        (
            "zip",
            None,
            """Zip code (USA) or "zip code,country code" for
            other countries. E.g. 12345,NZ. Takes precedence over
            coordinates.""",
        ),
        (
            "coordinates",
            None,
            """Dictionary containing latitude and longitude
               Example: coordinates={"longitude": "77.22",
                                     "latitude": "28.67"}""",
        ),
        (
            "format",
            "{location_city}: {main_temp} °{units_temperature} {humidity}% {weather_details}",
            "Display format",
        ),
        ("metric", True, "True to use metric/C, False to use imperial/F"),
        (
            "dateformat",
            "%Y-%m-%d ",
            """Format for dates, defaults to ISO.
            For details see: https://docs.python.org/3/library/time.html#time.strftime""",
        ),
        (
            "timeformat",
            "%H:%M",
            """Format for times, defaults to ISO.
            For details see: https://docs.python.org/3/library/time.html#time.strftime""",
        ),
        (
            "language",
            "en",
            """Language of response. List of languages supported can
         be seen at: https://openweathermap.org/current under
         Multilingual support""",
        ),
        (
            "weather_symbols",
            dict(),
            "Dictionary of weather symbols. Can be used to override default symbols.",
        ),
    ]

    def __init__(self, **config):
        GenPollUrl.__init__(self, **config)
        self.add_defaults(OpenWeather.defaults)
        self.symbols.update(self.weather_symbols)

    @property
    def url(self):
        if not self.cityid and not self.location and not self.zip and not self.coordinates:
            return None

        params = {
            "appid": self.app_key or DEFAULT_APP_ID,
            "units": "metric" if self.metric else "imperial",
        }
        if self.cityid:
            params["id"] = self.cityid
        elif self.location:
            params["q"] = self.location
        elif self.zip:
            params["zip"] = self.zip
        elif self.coordinates:
            params["lat"] = self.coordinates["latitude"]
            params["lon"] = self.coordinates["longitude"]

        if self.language:
            params["lang"] = self.language

        return QUERY_URL + urlencode(params)

    def parse(self, response):
        try:
            rp = _OpenWeatherResponseParser(response, self.dateformat, self.timeformat)
        except OpenWeatherResponseError as e:
            return f"Error {e.resp_code}"

        data = rp.data
        data["units_temperature"] = "C" if self.metric else "F"
        data["units_wind_speed"] = "Km/h" if self.metric else "m/h"
        data["icon"] = self.symbols.get(data["weather_0_icon"], self.symbols["Unknown"])

        return self.format.format(**data)
